#
# The MIT License (MIT)
# Copyright (C) 2014 SUSE LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Module for managing XFS file systems.
"""


import logging
import os
import re
import time

import salt.utils.data
import salt.utils.files
import salt.utils.path
import salt.utils.platform
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only work on POSIX-like systems
    """
    return not salt.utils.platform.is_windows() and __grains__.get("kernel") == "Linux"


def _verify_run(out, cmd=None):
    """
    Crash to the log if command execution was not successful.
    """
    if out.get("retcode", 0) and out["stderr"]:
        if cmd:
            log.debug('Command: "%s"', cmd)

        log.debug("Return code: %s", out.get("retcode"))
        log.debug("Error output:\n%s", out.get("stderr", "N/A"))

        raise CommandExecutionError(out["stderr"])


def _xfs_info_get_kv(serialized):
    """
    Parse one line of the XFS info output.
    """
    # No need to know sub-elements here
    if serialized.startswith("="):
        serialized = serialized[1:].strip()

    serialized = serialized.replace(" = ", "=*** ").replace(" =", "=")

    # Keywords has no spaces, values do
    opt = []
    for tkn in serialized.split(" "):
        if not opt or "=" in tkn:
            opt.append(tkn)
        else:
            opt[len(opt) - 1] = opt[len(opt) - 1] + " " + tkn

    # Preserve ordering
    return [tuple(items.split("=")) for items in opt]


def _parse_xfs_info(data):
    """
    Parse output from "xfs_info" or "xfs_growfs -n".
    """
    ret = {}
    spr = re.compile(r"\s+")
    entry = None
    for line in [spr.sub(" ", l).strip().replace(", ", " ") for l in data.split("\n")]:
        if not line or "=" not in line:
            continue
        nfo = _xfs_info_get_kv(line)
        if not line.startswith("="):
            entry = nfo.pop(0)
            ret[entry[0]] = {"section": entry[(entry[1] != "***" and 1 or 0)]}
        ret[entry[0]].update(dict(nfo))

    return ret


def info(device):
    """
    Get filesystem geometry information.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.info /dev/sda1
    """
    out = __salt__["cmd.run_all"](f"xfs_info {device}")
    if out.get("stderr"):
        raise CommandExecutionError(out["stderr"].replace("xfs_info:", "").strip())
    return _parse_xfs_info(out["stdout"])


def _xfsdump_output(data):
    """
    Parse CLI output of the xfsdump utility.
    """
    out = {}
    summary = []
    summary_block = False

    for line in [l.strip() for l in data.split("\n") if l.strip()]:
        line = re.sub("^xfsdump: ", "", line)
        if line.startswith("session id:"):
            out["Session ID"] = line.split(" ")[-1]
        elif line.startswith("session label:"):
            out["Session label"] = re.sub("^session label: ", "", line)
        elif line.startswith("media file size"):
            out["Media size"] = re.sub(r"^media file size\s+", "", line)
        elif line.startswith("dump complete:"):
            out["Dump complete"] = re.sub(r"^dump complete:\s+", "", line)
        elif line.startswith("Dump Status:"):
            out["Status"] = re.sub(r"^Dump Status:\s+", "", line)
        elif line.startswith("Dump Summary:"):
            summary_block = True
            continue

        if line.startswith(" ") and summary_block:
            summary.append(line.strip())
        elif not line.startswith(" ") and summary_block:
            summary_block = False

    if summary:
        out["Summary"] = " ".join(summary)

    return out


def dump(device, destination, level=0, label=None, noerase=None):
    """
    Dump filesystem device to the media (file, tape etc).

    Required parameters:

    * **device**: XFS device, content of which to be dumped.
    * **destination**: Specifies a dump destination.

    Valid options are:

    * **label**: Label of the dump. Otherwise automatically generated label is used.
    * **level**: Specifies a dump level of 0 to 9.
    * **noerase**: Pre-erase media.

    Other options are not used in order to let ``xfsdump`` use its default
    values, as they are most optimal. See the ``xfsdump(8)`` manpage for
    a more complete description of these options.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.dump /dev/sda1 /detination/on/the/client
        salt '*' xfs.dump /dev/sda1 /detination/on/the/client label='Company accountancy'
        salt '*' xfs.dump /dev/sda1 /detination/on/the/client noerase=True
    """
    if not salt.utils.path.which("xfsdump"):
        raise CommandExecutionError('Utility "xfsdump" has to be installed or missing.')

    label = (
        label
        and label
        or time.strftime(
            f'XFS dump for "{device}" of %Y.%m.%d, %H:%M', time.localtime()
        ).replace("'", '"')
    )
    cmd = ["xfsdump"]
    cmd.append("-F")  # Force
    if not noerase:
        cmd.append("-E")  # pre-erase
    cmd.append(f"-L '{label}'")  # Label
    cmd.append(f"-l {level}")  # Dump level
    cmd.append(f"-f {destination}")  # Media destination
    cmd.append(device)  # Device

    cmd = " ".join(cmd)
    out = __salt__["cmd.run_all"](cmd)
    _verify_run(out, cmd=cmd)

    return _xfsdump_output(out["stdout"])


def _xr_to_keyset(line):
    """
    Parse xfsrestore output keyset elements.
    """
    tkns = [elm for elm in line.strip().split(":", 1) if elm]
    if len(tkns) == 1:
        return f"'{tkns[0]}': "
    else:
        key, val = tkns
        return f"'{key.strip()}': '{val.strip()}',"


def _xfs_inventory_output(out):
    """
    Transform xfsrestore inventory data output to a Python dict source and evaluate it.
    """
    data = []
    out = [line for line in out.split("\n") if line.strip()]

    # No inventory yet
    if len(out) == 1 and "restore status" in out[0].lower():
        return {"restore_status": out[0]}

    ident = 0
    data.append("{")
    for line in out[:-1]:
        if len([elm for elm in line.strip().split(":") if elm]) == 1:
            n_ident = len(re.sub("[^\t]", "", line))
            if ident > n_ident:
                for step in range(ident):
                    data.append("},")
            ident = n_ident
            data.append(_xr_to_keyset(line))
            data.append("{")
        else:
            data.append(_xr_to_keyset(line))
    for step in range(ident + 1):
        data.append("},")
    data.append("},")

    # We are evaling into a python dict, a json load
    # would be safer
    data = eval("\n".join(data))[0]  # pylint: disable=W0123
    data["restore_status"] = out[-1]

    return data


def inventory():
    """
    Display XFS dump inventory without restoration.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.inventory
    """
    out = __salt__["cmd.run_all"]("xfsrestore -I")
    _verify_run(out)

    return _xfs_inventory_output(out["stdout"])


def _xfs_prune_output(out, uuid):
    """
    Parse prune output.
    """
    data = {}
    cnt = []
    cutpoint = False
    for line in [l.strip() for l in out.split("\n") if l]:
        if line.startswith("-"):
            if cutpoint:
                break
            else:
                cutpoint = True
                continue

        if cutpoint:
            cnt.append(line)

    for kset in [e for e in cnt[1:] if ":" in e]:
        key, val = (t.strip() for t in kset.split(":", 1))
        data[key.lower().replace(" ", "_")] = val

    return data.get("uuid") == uuid and data or {}


def prune_dump(sessionid):
    """
    Prunes the dump session identified by the given session id.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.prune_dump b74a3586-e52e-4a4a-8775-c3334fa8ea2c

    """
    out = __salt__["cmd.run_all"](f"xfsinvutil -s {sessionid} -F")
    _verify_run(out)

    data = _xfs_prune_output(out["stdout"], sessionid)
    if data:
        return data

    raise CommandExecutionError(f'Session UUID "{sessionid}" was not found.')


def _blkid_output(out):
    """
    Parse blkid output.
    """

    def flt(data):
        return [el for el in data if el.strip()]

    data = {}
    for dev_meta in flt(out.split("\n\n")):
        dev = {}
        for items in flt(dev_meta.strip().split("\n")):
            key, val = items.split("=", 1)
            dev[key.lower()] = val
        if dev.pop("type", None) == "xfs":
            dev["label"] = dev.get("label")
            data[dev.pop("devname")] = dev

    mounts = _get_mounts()
    for device in mounts:
        if data.get(device):
            data[device].update(mounts[device])

    return data


def devices():
    """
    Get known XFS formatted devices on the system.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.devices
    """
    out = __salt__["cmd.run_all"]("blkid -o export")
    _verify_run(out)

    return _blkid_output(out["stdout"])


def _xfs_estimate_output(out):
    """
    Parse xfs_estimate output.
    """
    spc = re.compile(r"\s+")
    data = {}
    for line in [l for l in out.split("\n") if l.strip()][1:]:
        directory, bsize, blocks, megabytes, logsize = spc.sub(" ", line).split(" ")
        data[directory] = {
            "block _size": bsize,
            "blocks": blocks,
            "megabytes": megabytes,
            "logsize": logsize,
        }

    return data


def estimate(path):
    """
    Estimate the space that an XFS filesystem will take.
    For each directory estimate the space that directory would take
    if it were copied to an XFS filesystem.
    Estimation does not cross mount points.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.estimate /path/to/file
        salt '*' xfs.estimate /path/to/dir/*
    """
    if not os.path.exists(path):
        raise CommandExecutionError(f'Path "{path}" was not found.')

    out = __salt__["cmd.run_all"](f"xfs_estimate -v {path}")
    _verify_run(out)

    return _xfs_estimate_output(out["stdout"])


def mkfs(
    device,
    label=None,
    ssize=None,
    noforce=None,
    bso=None,
    gmo=None,
    ino=None,
    lso=None,
    rso=None,
    nmo=None,
    dso=None,
):
    """
    Create a file system on the specified device. By default wipes out with force.

    General options:

    * **label**: Specify volume label.
    * **ssize**: Specify the fundamental sector size of the filesystem.
    * **noforce**: Do not force create filesystem, if disk is already formatted.

    Filesystem geometry options:

    * **bso**: Block size options.
    * **gmo**: Global metadata options.
    * **dso**: Data section options. These options specify the location, size,
               and other parameters of the data section of the filesystem.
    * **ino**: Inode options to specify the inode size of the filesystem, and other inode allocation parameters.
    * **lso**: Log section options.
    * **nmo**: Naming options.
    * **rso**: Realtime section options.

    See the ``mkfs.xfs(8)`` manpage for a more complete description of corresponding options description.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.mkfs /dev/sda1
        salt '*' xfs.mkfs /dev/sda1 dso='su=32k,sw=6' noforce=True
        salt '*' xfs.mkfs /dev/sda1 dso='su=32k,sw=6' lso='logdev=/dev/sda2,size=10000b'
    """

    def getopts(args):
        return dict(
            (args and ("=" in args) and args or None)
            and [kw.split("=") for kw in args.split(",")]
            or []
        )

    cmd = ["mkfs.xfs"]
    if label:
        cmd.append("-L")
        cmd.append(f"'{label}'")

    if ssize:
        cmd.append("-s")
        cmd.append(ssize)

    for switch, opts in [
        ("-b", bso),
        ("-m", gmo),
        ("-n", nmo),
        ("-i", ino),
        ("-d", dso),
        ("-l", lso),
        ("-r", rso),
    ]:
        try:
            if getopts(opts):
                cmd.append(switch)
                cmd.append(opts)
        except Exception:  # pylint: disable=broad-except
            raise CommandExecutionError(
                f'Wrong parameters "{opts}" for option "{switch}"'
            )

    if not noforce:
        cmd.append("-f")
    cmd.append(device)

    cmd = " ".join(cmd)
    out = __salt__["cmd.run_all"](cmd)
    _verify_run(out, cmd=cmd)

    return _parse_xfs_info(out["stdout"])


def modify(device, label=None, lazy_counting=None, uuid=None):
    """
    Modify parameters of an XFS filesystem.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.modify /dev/sda1 label='My backup' lazy_counting=False
        salt '*' xfs.modify /dev/sda1 uuid=False
        salt '*' xfs.modify /dev/sda1 uuid=True
    """
    if not label and lazy_counting is None and uuid is None:
        raise CommandExecutionError(
            f'Nothing specified for modification for "{device}" device'
        )

    cmd = ["xfs_admin"]
    if label:
        cmd.append("-L")
        cmd.append(f"'{label}'")

    if lazy_counting is False:
        cmd.append("-c")
        cmd.append("0")
    elif lazy_counting:
        cmd.append("-c")
        cmd.append("1")

    if uuid is False:
        cmd.append("-U")
        cmd.append("nil")
    elif uuid:
        cmd.append("-U")
        cmd.append("generate")
    cmd.append(device)

    cmd = " ".join(cmd)
    _verify_run(__salt__["cmd.run_all"](cmd), cmd=cmd)

    out = __salt__["cmd.run_all"](f"blkid -o export {device}")
    _verify_run(out)

    return _blkid_output(out["stdout"])


def _get_mounts():
    """
    List mounted filesystems.
    """
    mounts = {}
    with salt.utils.files.fopen("/proc/mounts") as fhr:
        for line in salt.utils.data.decode(fhr.readlines()):
            device, mntpnt, fstype, options, fs_freq, fs_passno = line.strip().split(
                " "
            )
            if fstype != "xfs":
                continue
            mounts[device] = {
                "mount_point": mntpnt,
                "options": options.split(","),
            }

    return mounts


def defragment(device):
    """
    Defragment mounted XFS filesystem.
    In order to mount a filesystem, device should be properly mounted and writable.

    CLI Example:

    .. code-block:: bash

        salt '*' xfs.defragment /dev/sda1
    """
    if device == "/":
        raise CommandExecutionError("Root is not a device.")

    if not _get_mounts().get(device):
        raise CommandExecutionError(f'Device "{device}" is not mounted')

    out = __salt__["cmd.run_all"](f"xfs_fsr {device}")
    _verify_run(out)

    return {"log": out["stdout"]}



"""
tests for host state
"""

from tests.support.case import ModuleCase


class HandleOrderTest(ModuleCase):
    """
    Validate that ordering works correctly
    """

    def test_handle_iorder(self):
        """
        Test the error with multiple states of the same type
        """
        ret = self.run_function("state.show_low_sls", mods="issue-7649-handle-iorder")

        sorted_chunks = [
            chunk["name"] for chunk in sorted(ret, key=lambda c: c.get("order"))
        ]

        expected = ["./configure", "make", "make install"]
        self.assertEqual(expected, sorted_chunks)



import logging
import os

import salt.modules.cmdmod
import salt.utils.json
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

__virtualname__ = "ansible"

log = logging.getLogger(__name__)


def __virtual__():
    if salt.utils.path.which("ansible-inventory"):
        return __virtualname__
    return (False, "Install `ansible` to use inventory")


def targets(inventory="/etc/ansible/hosts", yaml=False, export=False):
    """
    Return the targets from the ansible inventory_file
    Default: /etc/salt/roster
    """
    if not os.path.isfile(inventory):
        raise CommandExecutionError(f"Inventory file not found: {inventory}")
    if not os.path.isabs(inventory):
        raise CommandExecutionError("Path to inventory file must be an absolute path")

    extra_cmd = []
    if export:
        extra_cmd.append("--export")
    if yaml:
        extra_cmd.append("--yaml")
    inv = salt.modules.cmdmod.run(
        "ansible-inventory -i {} --list {}".format(inventory, " ".join(extra_cmd)),
        env={"ANSIBLE_DEPRECATION_WARNINGS": "0"},
        reset_system_locale=False,
    )
    if yaml:
        return salt.utils.stringutils.to_str(inv)
    else:
        try:
            return salt.utils.json.loads(salt.utils.stringutils.to_str(inv))
        except ValueError:
            raise CommandExecutionError(f"Error processing the inventory: {inv}")



import logging

import pytest

import salt.modules.etcd_mod as etcd_mod
import salt.states.etcd_mod as etcd_state
from salt.utils.etcd_util import get_conn
from tests.support.pytest.etcd import *  # pylint: disable=wildcard-import,unused-wildcard-import

docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_if_binaries_missing("docker", "dockerd", check_all=False),
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        etcd_state: {
            "__salt__": {
                "etcd.get": etcd_mod.get_,
                "etcd.set": etcd_mod.set_,
                "etcd.rm": etcd_mod.rm_,
            },
        },
        etcd_mod: {
            "__opts__": minion_opts,
            "__utils__": {
                "etcd_util.get_conn": get_conn,
            },
        },
    }


@pytest.fixture(scope="module")
def minion_config_overrides(etcd_profile):
    return etcd_profile


@pytest.fixture(scope="module")
def etcd_client(minion_opts, profile_name):
    return get_conn(minion_opts, profile=profile_name)


@pytest.fixture(scope="module")
def prefix():
    return "/salt/states/test"


@pytest.fixture(autouse=True)
def cleanup_prefixed_entries(etcd_client, prefix):
    """
    Cleanup after each test to ensure a consistent starting state.
    """
    try:
        assert etcd_client.get(prefix, recurse=True) is None
        yield
    finally:
        etcd_client.delete(prefix, recurse=True)


def test_basic_operations(subtests, profile_name, prefix, etcd_version):
    """
    Test basic CRUD operations
    """
    with subtests.test("Removing a non-existent key should not explode"):
        expected = {
            "name": f"{prefix}/2/3",
            "comment": "Key does not exist",
            "result": True,
            "changes": {},
        }
        assert etcd_state.rm(f"{prefix}/2/3", profile=profile_name) == expected

    with subtests.test("We should be able to set a value"):
        expected = {
            "name": f"{prefix}/1",
            "comment": "New key created",
            "result": True,
            "changes": {f"{prefix}/1": "one"},
        }
        assert etcd_state.set_(f"{prefix}/1", "one", profile=profile_name) == expected

    with subtests.test(
        "We should be able to create an empty directory and set values in it"
    ):
        if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode):
            expected = {
                "name": f"{prefix}/2",
                "comment": "New directory created",
                "result": True,
                "changes": {f"{prefix}/2": "Created"},
            }
            assert etcd_state.directory(f"{prefix}/2", profile=profile_name) == expected

        expected = {
            "name": f"{prefix}/2/3",
            "comment": "New key created",
            "result": True,
            "changes": {f"{prefix}/2/3": "two-three"},
        }
        assert (
            etcd_state.set_(f"{prefix}/2/3", "two-three", profile=profile_name)
            == expected
        )

    with subtests.test("We should be able to remove an existing key"):
        expected = {
            "name": f"{prefix}/2/3",
            "comment": "Key removed",
            "result": True,
            "changes": {f"{prefix}/2/3": "Deleted"},
        }
        assert etcd_state.rm(f"{prefix}/2/3", profile=profile_name) == expected


def test_with_missing_profile(subtests, prefix, etcd_version, etcd_port):
    """
    Test the correct response when the profile is missing and we can't connect
    """
    if etcd_version in (EtcdVersion.v2, EtcdVersion.v3_v2_mode) and etcd_port != 2379:
        # Only need to run this once
        with subtests.test("Test no profile and bad connection in set_"):
            ret = etcd_state.set_(f"{prefix}/1", "one")
            assert not ret["result"]
            assert ret["comment"] == etcd_state.NO_PROFILE_MSG

        with subtests.test("Test no profile and bad connection in directory"):
            ret = etcd_state.directory(f"{prefix}/2")
            assert not ret["result"]
            assert ret["comment"] == etcd_state.NO_PROFILE_MSG

        with subtests.test("Test no profile and bad connection in rm"):
            ret = etcd_state.rm(f"{prefix}/2/3")
            assert not ret["result"]
            assert ret["comment"] == etcd_state.NO_PROFILE_MSG



# pylint: disable=wrong-import-order,wrong-import-position,3rd-party-local-module-not-gated
# pylint: disable=redefined-outer-name,invalid-name,3rd-party-module-not-gated


import logging
import os
import pathlib
import re
import shutil
import stat
import sys
from functools import lru_cache
from unittest import TestCase  # pylint: disable=blacklisted-module

import _pytest.logging
import _pytest.skipping
import more_itertools
import pytest
import pytestskipmarkers

import salt
import salt._logging
import salt._logging.mixins
import salt.config
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.win_functions
from salt.serializers import yaml
from salt.utils.immutabletypes import freeze
from tests.support.helpers import (
    PRE_PYTEST_SKIP_OR_NOT,
    PRE_PYTEST_SKIP_REASON,
    get_virtualenv_binary_path,
)
from tests.support.pytest.helpers import *  # pylint: disable=unused-wildcard-import,wildcard-import
from tests.support.runtests import RUNTIME_VARS
from tests.support.sminion import check_required_sminion_attributes, create_sminion

TESTS_DIR = pathlib.Path(__file__).resolve().parent
PYTESTS_DIR = TESTS_DIR / "pytests"
CODE_DIR = TESTS_DIR.parent

# Change to code checkout directory
os.chdir(str(CODE_DIR))

# Make sure the current directory is the first item in sys.path
if str(CODE_DIR) in sys.path:
    sys.path.remove(str(CODE_DIR))
if os.environ.get("ONEDIR_TESTRUN", "0") == "0":
    sys.path.insert(0, str(CODE_DIR))

os.environ["REPO_ROOT_DIR"] = str(CODE_DIR)

# Coverage
if "COVERAGE_PROCESS_START" in os.environ:
    MAYBE_RUN_COVERAGE = True
    COVERAGERC_FILE = os.environ["COVERAGE_PROCESS_START"]
else:
    COVERAGERC_FILE = str(CODE_DIR / ".coveragerc")
    MAYBE_RUN_COVERAGE = (
        sys.argv[0].endswith("pytest.py") or "_COVERAGE_RCFILE" in os.environ
    )
    if MAYBE_RUN_COVERAGE:
        # Flag coverage to track suprocesses by pointing it to the right .coveragerc file
        os.environ["COVERAGE_PROCESS_START"] = str(COVERAGERC_FILE)

# Variable defining a FIPS test run or not
FIPS_TESTRUN = os.environ.get("FIPS_TESTRUN", "0") == "1"

# Define the pytest plugins we rely on
pytest_plugins = ["helpers_namespace"]

# Define where not to collect tests from
collect_ignore = ["setup.py"]


# Patch PyTest logging handlers
class LogCaptureHandler(
    salt._logging.mixins.ExcInfoOnLogLevelFormatMixin, _pytest.logging.LogCaptureHandler
):
    """
    Subclassing PyTest's LogCaptureHandler in order to add the
    exc_info_on_loglevel functionality and actually make it a NullHandler,
    it's only used to print log messages emmited during tests, which we
    have explicitly disabled in pytest.ini
    """


_pytest.logging.LogCaptureHandler = LogCaptureHandler


class LiveLoggingStreamHandler(
    salt._logging.mixins.ExcInfoOnLogLevelFormatMixin,
    _pytest.logging._LiveLoggingStreamHandler,
):
    """
    Subclassing PyTest's LiveLoggingStreamHandler in order to add the
    exc_info_on_loglevel functionality.
    """


_pytest.logging._LiveLoggingStreamHandler = LiveLoggingStreamHandler

# Reset logging root handlers
for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)
    handler.close()


# Reset the root logger to its default level(because salt changed it)
logging.root.setLevel(logging.WARNING)

log = logging.getLogger("salt.testsuite")


# ----- CLI Options Setup ------------------------------------------------------------------------------------------->
def pytest_addoption(parser):
    """
    register argparse-style options and ini-style config values.
    """
    test_selection_group = parser.getgroup("Tests Selection")
    test_selection_group.addoption(
        "--from-filenames",
        default=None,
        help=(
            "Pass a comma-separated list of file paths, and any test module which"
            " corresponds to the specified file(s) will run. For example, if 'setup.py'"
            " was passed, then the corresponding test files defined in"
            " 'tests/filename_map.yml' would run. Absolute paths are assumed to be"
            " files containing relative paths, one per line. Providing the paths in a"
            " file can help get around shell character limits when the list of files is"
            " long."
        ),
    )
    # Add deprecated CLI flag until we completely switch to PyTest
    test_selection_group.addoption(
        "--names-file", default=None, help="Deprecated option"
    )
    test_selection_group.addoption(
        "--transport",
        default="zeromq",
        choices=("zeromq", "tcp"),
        help=(
            "Select which transport to run the integration tests with, zeromq or tcp."
            " Default: %(default)s"
        ),
    )
    test_selection_group.addoption(
        "--ssh",
        "--ssh-tests",
        dest="ssh",
        action="store_true",
        default=False,
        help=(
            "Run salt-ssh tests. These tests will spin up a temporary "
            "SSH server on your machine. In certain environments, this "
            "may be insecure! Default: False"
        ),
    )
    test_selection_group.addoption(
        "--no-fast",
        "--no-fast-tests",
        dest="fast",
        action="store_true",
        default=False,
        help="Don't run salt-fast tests. Default: %(default)s",
    )
    test_selection_group.addoption(
        "--run-slow",
        "--slow",
        "--slow-tests",
        dest="slow",
        action="store_true",
        default=False,
        help="Run slow tests. Default: %(default)s",
    )
    test_selection_group.addoption(
        "--core",
        "--core-tests",
        dest="core",
        action="store_true",
        default=False,
        help=(
            "Run salt-core tests. These tests test the engine of salt! "
            "Default: %(default)s"
        ),
    )
    test_selection_group.addoption(
        "--flaky",
        "--flaky-jail",
        dest="flaky",
        action="store_true",
        default=False,
        help=(
            "Run salt-flaky jail tests. These tests are in jail for being flaky! "
            "One day they will be made not flaky."
            "Default: %(default)s"
        ),
    )
    test_selection_group.addoption(
        "--proxy",
        "--proxy-tests",
        dest="proxy",
        action="store_true",
        default=False,
        help="Run proxy tests (DEPRECATED)",
    )

    output_options_group = parser.getgroup("Output Options")
    output_options_group.addoption(
        "--output-columns",
        default=80,
        type=int,
        help="Number of maximum columns to use on the output",
    )
    output_options_group.addoption(
        "--no-colors",
        "--no-colours",
        default=False,
        action="store_true",
        help="Disable colour printing.",
    )

    # ----- Test Groups --------------------------------------------------------------------------------------------->
    # This will allow running the tests in chunks
    test_selection_group.addoption(
        "--test-group-count",
        dest="test-group-count",
        type=int,
        help="The number of groups to split the tests into",
    )
    test_selection_group.addoption(
        "--test-group",
        dest="test-group",
        type=int,
        help="The group of tests that should be executed",
    )
    # <---- Test Groups ----------------------------------------------------------------------------------------------


# <---- CLI Options Setup --------------------------------------------------------------------------------------------


# ----- Register Markers -------------------------------------------------------------------------------------------->
@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    """
    called after command line options have been parsed
    and all plugins and initial conftest files been loaded.
    """
    try:
        assert config._onedir_check_complete
        return
    except AttributeError:
        if os.environ.get("ONEDIR_TESTRUN", "0") == "1":
            if pathlib.Path(salt.__file__).parent == CODE_DIR / "salt":
                raise pytest.UsageError(
                    "Apparently running the test suite against the onedir build "
                    "of salt, however, the imported salt package is pointing to "
                    "the respository checkout instead of the onedir package.\n\n"
                    f"  * sys.path: {sys.path}"
                )
        config._onedir_check_complete = True

    for dirname in CODE_DIR.iterdir():
        if not dirname.is_dir():
            continue
        if dirname != TESTS_DIR:
            config.addinivalue_line("norecursedirs", str(CODE_DIR / dirname))
    config.addinivalue_line(
        "norecursedirs",
        str(TESTS_DIR / "unit" / "modules" / "inspectlib" / "tree_test"),
    )

    # Expose the markers we use to pytest CLI
    config.addinivalue_line(
        "markers",
        "requires_salt_modules(*required_module_names): Skip if at least one module is"
        " not available.",
    )
    config.addinivalue_line(
        "markers",
        "requires_salt_states(*required_state_names): Skip if at least one state module"
        " is not available.",
    )
    config.addinivalue_line(
        "markers", "windows_whitelisted: Mark test as whitelisted to run under Windows"
    )
    config.addinivalue_line(
        "markers", "requires_sshd_server: Mark test that require an SSH server running"
    )
    config.addinivalue_line(
        "markers",
        "slow_test: Mark test as being slow. These tests are skipped by default unless"
        " `--run-slow` is passed",
    )
    config.addinivalue_line(
        "markers",
        "core_test: Mark test as being core. These tests are skipped by default unless"
        " `--core-tests` is passed",
    )
    config.addinivalue_line(
        "markers",
        "flaky_jail: Mark test as being jlaky. These tests are skipped by default unless"
        " `--flaky-jail` is passed",
    )
    config.addinivalue_line(
        "markers",
        "async_timeout: Timeout, in seconds, for asynchronous test functions(`async def`)",
    )
    config.addinivalue_line(
        "markers",
        "requires_random_entropy(minimum={}, timeout={}, skip=True): Mark test as "
        "requiring a minimum value of random entropy. In the case where the value is lower "
        "than the provided 'minimum', an attempt will be made to raise that value up until "
        "the provided 'timeout' minutes have passed, at which time, depending on the value "
        "of 'skip' the test will skip or fail.  For entropy poolsizes of 256 bits, the min "
        "is adjusted to 192.".format(
            EntropyGenerator.minimum_entropy, EntropyGenerator.max_minutes
        ),
    )
    config.addinivalue_line(
        "markers",
        "skip_initial_gh_actions_failure(skip=<boolean or callable, reason=None): Skip known test failures "
        "under the new GH Actions setup if the environment variable SKIP_INITIAL_GH_ACTIONS_FAILURES "
        "is equal to '1' and the 'skip' keyword argument is either `True` or it's a callable that "
        "when called returns `True`. If `skip` is a callable, it should accept a single argument "
        "'grains', which is the grains dictionary.",
    )
    config.addinivalue_line(
        "markers",
        "skip_initial_onedir_failure(skip=<boolean or callable, reason=None): Skip known test failures "
        "under the new onedir builds if the environment variable SKIP_INITIAL_ONEDIR_FAILURES "
        "is equal to '1' and the 'skip' keyword argument is either `True` or it's a callable that "
        "when called returns `True`. If `skip` is a callable, it should accept a single argument "
        "'grains', which is the grains dictionary.",
    )
    config.addinivalue_line(
        "markers",
        "timeout_unless_on_windows(*args, **kwargs): Apply the 'timeout' marker unless running "
        "on Windows.",
    )
    # "Flag" the slowTest decorator if we're skipping slow tests or not
    os.environ["SLOW_TESTS"] = str(config.getoption("--run-slow"))


# <---- Register Markers ---------------------------------------------------------------------------------------------


# ----- PyTest Tweaks ----------------------------------------------------------------------------------------------->
def set_max_open_files_limits(min_soft=3072, min_hard=4096):

    # Get current limits
    if salt.utils.platform.is_windows():
        import win32file

        prev_hard = win32file._getmaxstdio()
        prev_soft = 512
    else:
        import resource

        prev_soft, prev_hard = resource.getrlimit(resource.RLIMIT_NOFILE)

    # Check minimum required limits
    set_limits = False
    if prev_soft < min_soft:
        soft = min_soft
        set_limits = True
    else:
        soft = prev_soft

    if prev_hard < min_hard:
        hard = min_hard
        set_limits = True
    else:
        hard = prev_hard

    # Increase limits
    if set_limits:
        log.debug(
            " * Max open files settings is too low (soft: %s, hard: %s) for running the"
            " tests. Trying to raise the limits to soft: %s, hard: %s",
            prev_soft,
            prev_hard,
            soft,
            hard,
        )
        try:
            if salt.utils.platform.is_windows():
                hard = 2048 if hard > 2048 else hard
                win32file._setmaxstdio(hard)
            else:
                resource.setrlimit(resource.RLIMIT_NOFILE, (soft, hard))
        except Exception as err:  # pylint: disable=broad-except
            log.error(
                "Failed to raise the max open files settings -> %s. Please issue the"
                " following command on your console: 'ulimit -u %s'",
                err,
                soft,
            )
            exit(1)
    return soft, hard


def pytest_report_header(config):
    soft, hard = set_max_open_files_limits()
    transport = config.getoption("--transport")
    return f"max open files: soft={soft}; hard={hard}\nsalt-transport: {transport}"


def pytest_itemcollected(item):
    """We just collected a test item."""
    try:
        pathlib.Path(item.fspath.strpath).resolve().relative_to(PYTESTS_DIR)
        # Test is under tests/pytests
        if item.cls and issubclass(item.cls, TestCase):
            pytest.fail(
                "The tests under {0!r} MUST NOT use unittest's TestCase class or a"
                " subclass of it. Please move {1!r} outside of {0!r}".format(
                    str(PYTESTS_DIR.relative_to(CODE_DIR)), item.nodeid
                )
            )
    except ValueError:
        # Test is not under tests/pytests
        if not item.cls or (item.cls and not issubclass(item.cls, TestCase)):
            pytest.fail(
                "The test {!r} appears to be written for pytest but it's not under"
                " {!r}. Please move it there.".format(
                    item.nodeid,
                    str(PYTESTS_DIR.relative_to(CODE_DIR)),
                )
            )


@pytest.hookimpl(hookwrapper=True, trylast=True)
def pytest_collection_modifyitems(config, items):
    """
    called after collection has been performed, may filter or re-order
    the items in-place.

    :param _pytest.main.Session session: the pytest session object
    :param _pytest.config.Config config: pytest config object
    :param List[_pytest.nodes.Item] items: list of item objects
    """
    # Let PyTest or other plugins handle the initial collection
    yield
    groups_collection_modifyitems(config, items)
    from_filenames_collection_modifyitems(config, items)

    timeout_marker_tests_paths = (
        str(PYTESTS_DIR / "pkg"),
        str(PYTESTS_DIR / "scenarios"),
    )
    for item in items:
        marker = item.get_closest_marker("timeout_unless_on_windows")
        if marker is not None:
            if not salt.utils.platform.is_windows():
                # Apply the marker since we're not on windows
                marker_kwargs = marker.kwargs.copy()
                if "func_only" not in marker_kwargs:
                    # Default to counting only the test execution for the timeouts, ie,
                    # withough including the fixtures setup time towards the timeout.
                    marker_kwargs["func_only"] = True
                item.add_marker(pytest.mark.timeout(*marker.args, **marker_kwargs))
        else:
            if (
                not salt.utils.platform.is_windows()
                and not str(pathlib.Path(item.fspath).resolve()).startswith(
                    timeout_marker_tests_paths
                )
                and not item.get_closest_marker("timeout")
            ):
                # Let's apply the timeout marker on the test, if the marker
                # is not already applied
                # Default to counting only the test execution for the timeouts, ie,
                # withough including the fixtures setup time towards the timeout.
                item.add_marker(pytest.mark.timeout(90, func_only=True))


def pytest_markeval_namespace(config):
    """
    Called when constructing the globals dictionary used for evaluating string conditions in xfail/skipif markers.

    This is useful when the condition for a marker requires objects that are expensive or impossible to obtain during
    collection time, which is required by normal boolean conditions.

    :param config: The pytest config object.
    :returns: A dictionary of additional globals to add.
    """
    return {"grains": _grains_for_marker()}


# <---- PyTest Tweaks ------------------------------------------------------------------------------------------------


# ----- Test Setup -------------------------------------------------------------------------------------------------->
@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    """
    Fixtures injection based on markers or test skips based on CLI arguments
    """
    integration_utils_tests_path = str(TESTS_DIR / "integration" / "utils")
    if (
        str(item.fspath).startswith(integration_utils_tests_path)
        and PRE_PYTEST_SKIP_OR_NOT is True
    ):
        item._skipped_by_mark = True
        pytest.skip(PRE_PYTEST_SKIP_REASON)
    test_group_count = sum(
        bool(item.get_closest_marker(group))
        for group in ("core_test", "slow_test", "flaky_jail")
    )
    if item.get_closest_marker("core_test") and item.get_closest_marker("slow_test"):
        raise pytest.UsageError(
            "Tests can only be in one test group. ('core_test', 'slow_test')"
        )

    if item.get_closest_marker("flaky_jail"):
        if not item.config.getoption("--flaky-jail"):
            raise pytest.skip.Exception(
                "flaky jail tests are disabled, pass '--flaky-jail' to enable them.",
                _use_item_location=True,
            )
    else:
        if item.get_closest_marker("core_test"):
            if not item.config.getoption("--core-tests"):
                raise pytest.skip.Exception(
                    "Core tests are disabled, pass '--core-tests' to enable them.",
                    _use_item_location=True,
                )
        if item.get_closest_marker("slow_test"):
            if not item.config.getoption("--slow-tests"):
                raise pytest.skip.Exception(
                    "Slow tests are disabled, pass '--run-slow' to enable them.",
                    _use_item_location=True,
                )
        if test_group_count == 0 and item.config.getoption("--no-fast-tests"):
            raise pytest.skip.Exception(
                "Fast tests have been disabled by '--no-fast-tests'.",
                _use_item_location=True,
            )

    requires_sshd_server_marker = item.get_closest_marker("requires_sshd_server")
    if requires_sshd_server_marker is not None:
        if not item.config.getoption("--ssh-tests"):
            item._skipped_by_mark = True
            pytest.skip("SSH tests are disabled, pass '--ssh-tests' to enable them.")
        item.fixturenames.append("sshd_server")
        item.fixturenames.append("salt_ssh_roster_file")

    requires_salt_modules_marker = item.get_closest_marker("requires_salt_modules")
    if requires_salt_modules_marker is not None:
        required_salt_modules = requires_salt_modules_marker.args
        if len(required_salt_modules) == 1 and isinstance(
            required_salt_modules[0], (list, tuple, set)
        ):
            required_salt_modules = required_salt_modules[0]
        required_salt_modules = set(required_salt_modules)
        not_available_modules = check_required_sminion_attributes(
            "functions", required_salt_modules
        )

        if not_available_modules:
            item._skipped_by_mark = True
            if len(not_available_modules) == 1:
                pytest.skip(
                    "Salt module '{}' is not available".format(*not_available_modules)
                )
            pytest.skip(
                "Salt modules not available: {}".format(
                    ", ".join(not_available_modules)
                )
            )

    requires_salt_states_marker = item.get_closest_marker("requires_salt_states")
    if requires_salt_states_marker is not None:
        required_salt_states = requires_salt_states_marker.args
        if len(required_salt_states) == 1 and isinstance(
            required_salt_states[0], (list, tuple, set)
        ):
            required_salt_states = required_salt_states[0]
        required_salt_states = set(required_salt_states)
        not_available_states = check_required_sminion_attributes(
            "states", required_salt_states
        )

        if not_available_states:
            item._skipped_by_mark = True
            if len(not_available_states) == 1:
                pytest.skip(
                    "Salt state module '{}' is not available".format(
                        *not_available_states
                    )
                )
            pytest.skip(
                "Salt state modules not available: {}".format(
                    ", ".join(not_available_states)
                )
            )

    skip_initial_gh_actions_failures_env_set = (
        os.environ.get("SKIP_INITIAL_GH_ACTIONS_FAILURES", "0") == "1"
    )
    skip_initial_gh_actions_failure_marker = item.get_closest_marker(
        "skip_initial_gh_actions_failure"
    )
    if (
        skip_initial_gh_actions_failure_marker is not None
        and skip_initial_gh_actions_failures_env_set
    ):
        if skip_initial_gh_actions_failure_marker.args:
            raise pytest.UsageError(
                "'skip_initial_gh_actions_failure' marker does not accept any arguments "
                "only keyword arguments."
            )
        kwargs = skip_initial_gh_actions_failure_marker.kwargs.copy()
        skip = kwargs.pop("skip", True)
        if skip and not callable(skip) and not isinstance(skip, bool):
            raise pytest.UsageError(
                "The 'skip' keyword argument to the 'skip_initial_gh_actions_failure' marker "
                "requires a boolean or callable, not '{}'.".format(type(skip))
            )
        reason = kwargs.pop("reason", None)
        if reason is None:
            reason = "Test skipped because it's a know GH Actions initial failure that needs to be fixed"
        if kwargs:
            raise pytest.UsageError(
                "'skip_initial_gh_actions_failure' marker does not accept any keyword arguments "
                "except 'skip' and 'reason'."
            )
        if skip and callable(skip):
            grains = _grains_for_marker()
            skip = skip(grains)

        if skip:
            raise pytest.skip.Exception(reason, _use_item_location=True)

    skip_initial_onedir_failures_env_set = (
        os.environ.get("SKIP_INITIAL_ONEDIR_FAILURES", "0") == "1"
    )
    skip_initial_onedir_failure_marker = item.get_closest_marker(
        "skip_initial_onedir_failure"
    )
    if (
        skip_initial_onedir_failure_marker is not None
        and skip_initial_onedir_failures_env_set
    ):
        if skip_initial_onedir_failure_marker.args:
            raise pytest.UsageError(
                "'skip_initial_onedir_failure' marker does not accept any arguments "
                "only keyword arguments."
            )
        kwargs = skip_initial_onedir_failure_marker.kwargs.copy()
        skip = kwargs.pop("skip", True)
        if skip and not callable(skip) and not isinstance(skip, bool):
            raise pytest.UsageError(
                "The 'skip' keyword argument to the 'skip_initial_onedir_failure' marker "
                "requires a boolean or callable, not '{}'.".format(type(skip))
            )
        reason = kwargs.pop("reason", None)
        if reason is None:
            reason = "Test skipped because it's a know GH Actions initial failure that needs to be fixed"
        if kwargs:
            raise pytest.UsageError(
                "'skip_initial_onedir_failure' marker does not accept any keyword arguments "
                "except 'skip' and 'reason'."
            )
        if skip and callable(skip):
            grains = _grains_for_marker()
            skip = skip(grains)

        if skip:
            raise pytest.skip.Exception(reason, _use_item_location=True)

    requires_random_entropy_marker = item.get_closest_marker("requires_random_entropy")
    if requires_random_entropy_marker is not None:
        if requires_random_entropy_marker.args:
            raise pytest.UsageError(
                "'requires_random_entropy' marker does not accept any arguments "
                "only keyword arguments."
            )
        kwargs = requires_random_entropy_marker.kwargs.copy()
        skip = kwargs.pop("skip", None)
        if skip and not isinstance(skip, bool):
            raise pytest.UsageError(
                "The 'skip' keyword argument to the 'requires_random_entropy' marker "
                "requires a boolean not '{}'.".format(type(skip))
            )
        minimum_entropy = kwargs.pop("minimum", None)
        if minimum_entropy is not None:
            if not isinstance(minimum_entropy, int):
                raise pytest.UsageError(
                    "The 'minimum' keyword argument to the 'requires_random_entropy' marker "
                    "must be an integer not '{}'.".format(type(minimum_entropy))
                )
            if minimum_entropy <= 0:
                raise pytest.UsageError(
                    "The 'minimum' keyword argument to the 'requires_random_entropy' marker "
                    "must be an positive integer not '{}'.".format(minimum_entropy)
                )
        max_minutes = kwargs.pop("timeout", None)
        if max_minutes is not None:
            if not isinstance(max_minutes, int):
                raise pytest.UsageError(
                    "The 'timeout' keyword argument to the 'requires_random_entropy' marker "
                    "must be an integer not '{}'.".format(type(max_minutes))
                )
            if max_minutes <= 0:
                raise pytest.UsageError(
                    "The 'timeout' keyword argument to the 'requires_random_entropy' marker "
                    "must be an positive integer not '{}'.".format(max_minutes)
                )
        if kwargs:
            raise pytest.UsageError(
                "Unsupported keyword arguments passed to the 'requires_random_entropy' "
                "marker: {}".format(", ".join(list(kwargs)))
            )
        entropy_generator = EntropyGenerator(
            minimum_entropy=minimum_entropy, max_minutes=max_minutes, skip=skip
        )
        entropy_generator.generate_entropy()

    if salt.utils.platform.is_windows():
        auto_whitelisted_paths = (
            str(TESTS_DIR / "unit"),
            str(PYTESTS_DIR / "unit"),
            str(PYTESTS_DIR / "pkg"),
        )
        if not str(pathlib.Path(item.fspath).resolve()).startswith(
            auto_whitelisted_paths
        ):
            # Unit tests are whitelisted on windows by default, so, we're only
            # after all other tests
            windows_whitelisted_marker = item.get_closest_marker("windows_whitelisted")
            if windows_whitelisted_marker is None:
                item._skipped_by_mark = True
                pytest.skip("Test is not whitelisted for Windows")


# <---- Test Setup ---------------------------------------------------------------------------------------------------


# ----- Test Groups Selection --------------------------------------------------------------------------------------->
def groups_collection_modifyitems(config, items):
    group_count = config.getoption("test-group-count")
    group_id = config.getoption("test-group")

    if not group_count or not group_id:
        # We're not selection tests using groups, don't do any filtering
        return

    if group_count == 1:
        # Just one group, don't do any filtering
        return

    terminal_reporter = config.pluginmanager.get_plugin("terminalreporter")

    if config.getoption("--last-failed") or config.getoption("--failed-first"):
        # This is a test failure rerun, applying test groups would break this
        terminal_reporter.write(
            "\nNot splitting collected tests into chunks since --lf/--last-failed or "
            "-ff/--failed-first was passed on the CLI.\n",
            yellow=True,
        )
        return

    total_items = len(items)

    # Devide into test groups
    test_groups = more_itertools.divide(group_count, items)
    # Pick the right group
    tests_in_group = list(test_groups.pop(group_id - 1))
    # The rest are deselected tests
    deselected = list(more_itertools.collapse(test_groups))
    # Sanity check
    assert len(tests_in_group) + len(deselected) == total_items
    # Replace all items in the list
    items[:] = tests_in_group
    if deselected:
        config.hook.pytest_deselected(items=deselected)

    terminal_reporter.write(
        f"Running test group #{group_id}(out of #{group_count}) ({len(items)} out of {total_items} tests)\n",
        yellow=True,
    )


# <---- Test Groups Selection ----------------------------------------------------------------------------------------


# ----- Fixtures Overrides ------------------------------------------------------------------------------------------>
@pytest.fixture(scope="session")
def salt_factories_default_root_dir(salt_factories_default_root_dir):
    """
    The root directory from where to base all salt-factories paths.

    For example, in a salt system installation, this would be ``/``.

    .. admonition:: Attention

        If `root_dir` is returned on the `salt_factories_config()` fixture
        dictionary, then that's the value used, and not the one returned by
        this fixture.
    """
    if os.environ.get("CI") and pytestskipmarkers.utils.platform.is_windows():
        tempdir = pathlib.Path(
            os.environ.get("RUNNER_TEMP", r"C:\Windows\Temp")
        ).resolve()
        return tempdir / "stsuite"

    return salt_factories_default_root_dir / "stsuite"


@pytest.fixture(scope="session")
def salt_factories_config():
    """
    Return a dictionary with the keyworkd arguments for FactoriesManager
    """
    if os.environ.get("JENKINS_URL") or os.environ.get("CI"):
        start_timeout = 120
    else:
        start_timeout = 60

    if os.environ.get("ONEDIR_TESTRUN", "0") == "1":
        code_dir = None
    else:
        code_dir = str(CODE_DIR)

    kwargs = {
        "code_dir": code_dir,
        "start_timeout": start_timeout,
        "inject_sitecustomize": MAYBE_RUN_COVERAGE,
    }
    if MAYBE_RUN_COVERAGE:
        kwargs["coverage_rc_path"] = str(COVERAGERC_FILE)
    else:
        kwargs["coverage_rc_path"] = None
    kwargs["coverage_db_path"] = os.environ.get("COVERAGE_FILE")
    return kwargs


@pytest.fixture
def tmpdir(tmpdir):
    raise pytest.UsageError(
        "The `tmpdir` fixture uses Pytest's `pypath` implementation which "
        "is getting deprecated in favor of `pathlib`. Please use the "
        "`tmp_path` fixture instead."
    )


# <---- Fixtures Overrides -------------------------------------------------------------------------------------------


# ----- Salt Factories ---------------------------------------------------------------------------------------------->
@pytest.fixture(scope="session")
def integration_files_dir(salt_factories):
    """
    Fixture which returns the salt integration files directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = salt_factories.root_dir / "integration-files"
    dirname.mkdir(exist_ok=True)
    for child in (PYTESTS_DIR / "integration" / "files").iterdir():
        destpath = dirname / child.name
        if child.is_dir():
            shutil.copytree(str(child), str(destpath), dirs_exist_ok=True)
        else:
            shutil.copyfile(str(child), str(destpath))
    return dirname


@pytest.fixture(scope="module")
def functional_files_dir(salt_factories):
    """
    Fixture which returns the salt functional files directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = salt_factories.root_dir / "functional-files"
    dirname.mkdir(exist_ok=True)
    for child in (PYTESTS_DIR / "functional" / "files").iterdir():
        if child.is_dir():
            shutil.copytree(str(child), str(dirname / child.name))
        else:
            shutil.copyfile(str(child), str(dirname / child.name))
    return dirname


@pytest.fixture(scope="session")
def state_tree_root_dir(integration_files_dir):
    """
    Fixture which returns the salt state tree root directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = integration_files_dir / "state-tree"
    dirname.mkdir(exist_ok=True)
    return dirname


@pytest.fixture(scope="session")
def pillar_tree_root_dir(integration_files_dir):
    """
    Fixture which returns the salt pillar tree root directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = integration_files_dir / "pillar-tree"
    dirname.mkdir(exist_ok=True)
    return dirname


@pytest.fixture(scope="session")
def base_env_state_tree_root_dir(state_tree_root_dir):
    """
    Fixture which returns the salt base environment state tree directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = state_tree_root_dir / "base"
    dirname.mkdir(exist_ok=True)
    RUNTIME_VARS.TMP_STATE_TREE = str(dirname.resolve())
    RUNTIME_VARS.TMP_BASEENV_STATE_TREE = RUNTIME_VARS.TMP_STATE_TREE
    return dirname


@pytest.fixture(scope="session")
def prod_env_state_tree_root_dir(state_tree_root_dir):
    """
    Fixture which returns the salt prod environment state tree directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = state_tree_root_dir / "prod"
    dirname.mkdir(exist_ok=True)
    RUNTIME_VARS.TMP_PRODENV_STATE_TREE = str(dirname.resolve())
    return dirname


@pytest.fixture(scope="session")
def base_env_pillar_tree_root_dir(pillar_tree_root_dir):
    """
    Fixture which returns the salt base environment pillar tree directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = pillar_tree_root_dir / "base"
    dirname.mkdir(exist_ok=True)
    RUNTIME_VARS.TMP_PILLAR_TREE = str(dirname.resolve())
    RUNTIME_VARS.TMP_BASEENV_PILLAR_TREE = RUNTIME_VARS.TMP_PILLAR_TREE
    return dirname


@pytest.fixture(scope="session")
def ext_pillar_file_tree_root_dir(pillar_tree_root_dir):
    """
    Fixture which returns the salt pillar file tree directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = pillar_tree_root_dir / "file-tree"
    dirname.mkdir(exist_ok=True)
    return dirname


@pytest.fixture(scope="session")
def prod_env_pillar_tree_root_dir(pillar_tree_root_dir):
    """
    Fixture which returns the salt prod environment pillar tree directory path.
    Creates the directory if it does not yet exist.
    """
    dirname = pillar_tree_root_dir / "prod"
    dirname.mkdir(exist_ok=True)
    RUNTIME_VARS.TMP_PRODENV_PILLAR_TREE = str(dirname.resolve())
    return dirname


@pytest.fixture(scope="session")
def salt_syndic_master_factory(
    request,
    salt_factories,
    base_env_state_tree_root_dir,
    base_env_pillar_tree_root_dir,
    prod_env_state_tree_root_dir,
    prod_env_pillar_tree_root_dir,
):
    root_dir = salt_factories.get_root_dir_for_daemon("syndic_master")
    conf_dir = root_dir / "conf"
    conf_dir.mkdir(exist_ok=True)

    with salt.utils.files.fopen(
        os.path.join(RUNTIME_VARS.CONF_DIR, "syndic_master")
    ) as rfh:
        config_defaults = yaml.deserialize(rfh.read())

        tests_known_hosts_file = str(root_dir / "salt_ssh_known_hosts")
        with salt.utils.files.fopen(tests_known_hosts_file, "w") as known_hosts:
            known_hosts.write("")

    config_defaults["root_dir"] = str(root_dir)
    config_defaults["known_hosts_file"] = tests_known_hosts_file
    config_defaults["syndic_master"] = "localhost"
    config_defaults["transport"] = request.config.getoption("--transport")

    config_overrides = {
        "log_level_logfile": "info",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    ext_pillar = []
    if salt.utils.platform.is_windows():
        ext_pillar.append(
            {"cmd_yaml": "type {}".format(os.path.join(RUNTIME_VARS.FILES, "ext.yaml"))}
        )
    else:
        ext_pillar.append(
            {"cmd_yaml": "cat {}".format(os.path.join(RUNTIME_VARS.FILES, "ext.yaml"))}
        )

    # We need to copy the extension modules into the new master root_dir or
    # it will be prefixed by it
    extension_modules_path = str(root_dir / "extension_modules")
    if not os.path.exists(extension_modules_path):
        shutil.copytree(
            os.path.join(RUNTIME_VARS.FILES, "extension_modules"),
            extension_modules_path,
        )

    # Copy the autosign_file to the new  master root_dir
    autosign_file_path = str(root_dir / "autosign_file")
    shutil.copyfile(
        os.path.join(RUNTIME_VARS.FILES, "autosign_file"), autosign_file_path
    )
    # all read, only owner write
    autosign_file_permissions = (
        stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR
    )
    os.chmod(autosign_file_path, autosign_file_permissions)

    config_overrides.update(
        {
            "ext_pillar": ext_pillar,
            "extension_modules": extension_modules_path,
            "file_roots": {
                "base": [
                    str(base_env_state_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "file", "base"),
                ],
                # Alternate root to test __env__ choices
                "prod": [
                    str(prod_env_state_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "file", "prod"),
                ],
            },
            "pillar_roots": {
                "base": [
                    str(base_env_pillar_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "pillar", "base"),
                ],
                "prod": [str(prod_env_pillar_tree_root_dir)],
            },
        }
    )

    factory = salt_factories.salt_master_daemon(
        "syndic_master",
        order_masters=True,
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    return factory


@pytest.fixture(scope="session")
def salt_syndic_factory(salt_factories, salt_syndic_master_factory):
    config_defaults = {"master": None, "minion": None, "syndic": None}
    with salt.utils.files.fopen(os.path.join(RUNTIME_VARS.CONF_DIR, "syndic")) as rfh:
        opts = yaml.deserialize(rfh.read())

        opts["hosts.file"] = os.path.join(RUNTIME_VARS.TMP, "hosts")
        opts["aliases.file"] = os.path.join(RUNTIME_VARS.TMP, "aliases")
        opts["transport"] = salt_syndic_master_factory.config["transport"]
        config_defaults["syndic"] = opts
    config_overrides = {"log_level_logfile": "info"}
    factory = salt_syndic_master_factory.salt_syndic_daemon(
        "syndic",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    return factory


@pytest.fixture(scope="session")
def salt_master_factory(
    salt_factories,
    salt_syndic_master_factory,
    base_env_state_tree_root_dir,
    base_env_pillar_tree_root_dir,
    prod_env_state_tree_root_dir,
    prod_env_pillar_tree_root_dir,
    ext_pillar_file_tree_root_dir,
    salt_api_account_factory,
):
    root_dir = salt_factories.get_root_dir_for_daemon("master")
    conf_dir = root_dir / "conf"
    conf_dir.mkdir(exist_ok=True)

    with salt.utils.files.fopen(os.path.join(RUNTIME_VARS.CONF_DIR, "master")) as rfh:
        config_defaults = yaml.deserialize(rfh.read())

        tests_known_hosts_file = str(root_dir / "salt_ssh_known_hosts")
        with salt.utils.files.fopen(tests_known_hosts_file, "w") as known_hosts:
            known_hosts.write("")

    config_defaults["root_dir"] = str(root_dir)
    config_defaults["known_hosts_file"] = tests_known_hosts_file
    config_defaults["syndic_master"] = "localhost"
    config_defaults["transport"] = salt_syndic_master_factory.config["transport"]

    config_overrides = {
        "log_level_logfile": "info",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    ext_pillar = []
    if salt.utils.platform.is_windows():
        ext_pillar.append(
            {"cmd_yaml": "type {}".format(os.path.join(RUNTIME_VARS.FILES, "ext.yaml"))}
        )
    else:
        ext_pillar.append(
            {"cmd_yaml": "cat {}".format(os.path.join(RUNTIME_VARS.FILES, "ext.yaml"))}
        )
    ext_pillar.append(
        {
            "file_tree": {
                "root_dir": str(ext_pillar_file_tree_root_dir),
                "follow_dir_links": False,
                "keep_newline": True,
            }
        }
    )
    config_overrides["pillar_opts"] = True
    config_overrides["external_auth"] = {
        "auto": {
            salt_api_account_factory.username: [
                "@wheel",
                "@runner",
                "test.*",
                "grains.*",
            ],
        }
    }

    # We need to copy the extension modules into the new master root_dir or
    # it will be prefixed by it
    extension_modules_path = str(root_dir / "extension_modules")
    if not os.path.exists(extension_modules_path):
        shutil.copytree(
            os.path.join(RUNTIME_VARS.FILES, "extension_modules"),
            extension_modules_path,
        )

    # Copy the autosign_file to the new  master root_dir
    autosign_file_path = str(root_dir / "autosign_file")
    shutil.copyfile(
        os.path.join(RUNTIME_VARS.FILES, "autosign_file"), autosign_file_path
    )
    # all read, only owner write
    autosign_file_permissions = (
        stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR
    )
    os.chmod(autosign_file_path, autosign_file_permissions)

    config_overrides.update(
        {
            "ext_pillar": ext_pillar,
            "extension_modules": extension_modules_path,
            "file_roots": {
                "base": [
                    str(base_env_state_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "file", "base"),
                ],
                # Alternate root to test __env__ choices
                "prod": [
                    str(prod_env_state_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "file", "prod"),
                ],
            },
            "pillar_roots": {
                "base": [
                    str(base_env_pillar_tree_root_dir),
                    os.path.join(RUNTIME_VARS.FILES, "pillar", "base"),
                ],
                "prod": [str(prod_env_pillar_tree_root_dir)],
            },
        }
    )

    # Let's copy over the test cloud config files and directories into the running master config directory
    for entry in os.listdir(RUNTIME_VARS.CONF_DIR):
        if not entry.startswith("cloud"):
            continue
        source = os.path.join(RUNTIME_VARS.CONF_DIR, entry)
        dest = str(conf_dir / entry)
        if os.path.isdir(source):
            shutil.copytree(source, dest)
        else:
            shutil.copyfile(source, dest)

    factory = salt_syndic_master_factory.salt_master_daemon(
        "master",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    return factory


@pytest.fixture(scope="session")
def salt_minion_factory(salt_master_factory):
    with salt.utils.files.fopen(os.path.join(RUNTIME_VARS.CONF_DIR, "minion")) as rfh:
        config_defaults = yaml.deserialize(rfh.read())
    config_defaults["hosts.file"] = os.path.join(RUNTIME_VARS.TMP, "hosts")
    config_defaults["aliases.file"] = os.path.join(RUNTIME_VARS.TMP, "aliases")
    config_defaults["transport"] = salt_master_factory.config["transport"]

    config_overrides = {
        "log_level_logfile": "info",
        "file_roots": salt_master_factory.config["file_roots"].copy(),
        "pillar_roots": salt_master_factory.config["pillar_roots"].copy(),
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }

    virtualenv_binary = get_virtualenv_binary_path()
    if virtualenv_binary:
        config_overrides["venv_bin"] = virtualenv_binary
    factory = salt_master_factory.salt_minion_daemon(
        "minion",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, salt_master_factory, factory.id
    )
    return factory


@pytest.fixture(scope="session")
def salt_sub_minion_factory(salt_master_factory):
    with salt.utils.files.fopen(
        os.path.join(RUNTIME_VARS.CONF_DIR, "sub_minion")
    ) as rfh:
        config_defaults = yaml.deserialize(rfh.read())
    config_defaults["hosts.file"] = os.path.join(RUNTIME_VARS.TMP, "hosts")
    config_defaults["aliases.file"] = os.path.join(RUNTIME_VARS.TMP, "aliases")
    config_defaults["transport"] = salt_master_factory.config["transport"]

    config_overrides = {
        "log_level_logfile": "info",
        "file_roots": salt_master_factory.config["file_roots"].copy(),
        "pillar_roots": salt_master_factory.config["pillar_roots"].copy(),
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }

    virtualenv_binary = get_virtualenv_binary_path()
    if virtualenv_binary:
        config_overrides["venv_bin"] = virtualenv_binary
    factory = salt_master_factory.salt_minion_daemon(
        "sub_minion",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, salt_master_factory, factory.id
    )
    return factory


@pytest.fixture(scope="session")
def salt_cli(salt_master_factory):
    return salt_master_factory.salt_cli()


@pytest.fixture(scope="session")
def salt_cp_cli(salt_master_factory):
    return salt_master_factory.salt_cp_cli()


@pytest.fixture(scope="session")
def salt_key_cli(salt_master_factory):
    return salt_master_factory.salt_key_cli()


@pytest.fixture(scope="session")
def salt_run_cli(salt_master_factory):
    return salt_master_factory.salt_run_cli()


@pytest.fixture(scope="session")
def salt_call_cli(salt_minion_factory):
    return salt_minion_factory.salt_call_cli()


@pytest.fixture(scope="session", autouse=True)
def bridge_pytest_and_runtests(
    salt_factories,
    salt_syndic_master_factory,
    salt_syndic_factory,
    salt_master_factory,
    salt_minion_factory,
    salt_sub_minion_factory,
    sshd_config_dir,
):
    # Make sure unittest2 uses the pytest generated configuration
    RUNTIME_VARS.RUNTIME_CONFIGS["master"] = freeze(salt_master_factory.config)
    RUNTIME_VARS.RUNTIME_CONFIGS["minion"] = freeze(salt_minion_factory.config)
    RUNTIME_VARS.RUNTIME_CONFIGS["sub_minion"] = freeze(salt_sub_minion_factory.config)
    RUNTIME_VARS.RUNTIME_CONFIGS["syndic_master"] = freeze(
        salt_syndic_master_factory.config
    )
    RUNTIME_VARS.RUNTIME_CONFIGS["syndic"] = freeze(salt_syndic_factory.config)
    RUNTIME_VARS.RUNTIME_CONFIGS["client_config"] = freeze(
        salt.config.client_config(salt_master_factory.config["conf_file"])
    )

    # Make sure unittest2 classes know their paths
    RUNTIME_VARS.TMP_ROOT_DIR = str(salt_factories.root_dir.resolve())
    RUNTIME_VARS.TMP_CONF_DIR = os.path.dirname(salt_master_factory.config["conf_file"])
    RUNTIME_VARS.TMP_MINION_CONF_DIR = os.path.dirname(
        salt_minion_factory.config["conf_file"]
    )
    RUNTIME_VARS.TMP_SUB_MINION_CONF_DIR = os.path.dirname(
        salt_sub_minion_factory.config["conf_file"]
    )
    RUNTIME_VARS.TMP_SYNDIC_MASTER_CONF_DIR = os.path.dirname(
        salt_syndic_master_factory.config["conf_file"]
    )
    RUNTIME_VARS.TMP_SYNDIC_MINION_CONF_DIR = os.path.dirname(
        salt_syndic_factory.config["conf_file"]
    )
    RUNTIME_VARS.TMP_SSH_CONF_DIR = str(sshd_config_dir)
    with reap_stray_processes():
        yield


@pytest.fixture(scope="session")
def sshd_config_dir(salt_factories):
    config_dir = salt_factories.get_root_dir_for_daemon("sshd")
    yield config_dir
    shutil.rmtree(str(config_dir), ignore_errors=True)


@pytest.fixture(scope="module")
def sshd_server(salt_factories, sshd_config_dir, salt_master, grains):
    sshd_config_dict = {
        "Protocol": "2",
        # Turn strict modes off so that we can operate in /tmp
        "StrictModes": "no",
        # Logging
        "SyslogFacility": "AUTH",
        "LogLevel": "INFO",
        # Authentication:
        "LoginGraceTime": "120",
        "PermitRootLogin": "without-password",
        "PubkeyAuthentication": "yes",
        # Don't read the user's ~/.rhosts and ~/.shosts files
        "IgnoreRhosts": "yes",
        "HostbasedAuthentication": "no",
        # To enable empty passwords, change to yes (NOT RECOMMENDED)
        "PermitEmptyPasswords": "no",
        # Change to yes to enable challenge-response passwords (beware issues with
        # some PAM modules and threads)
        "ChallengeResponseAuthentication": "no",
        # Change to no to disable tunnelled clear text passwords
        "PasswordAuthentication": "no",
        "X11Forwarding": "no",
        "X11DisplayOffset": "10",
        "PrintMotd": "no",
        "PrintLastLog": "yes",
        "TCPKeepAlive": "yes",
        "AcceptEnv": "LANG LC_*",
        "UsePAM": "yes",
    }
    sftp_server_paths = [
        # Common
        "/usr/lib/openssh/sftp-server",
        # CentOS Stream 9
        "/usr/libexec/openssh/sftp-server",
        # Arch Linux
        "/usr/lib/ssh/sftp-server",
        # Photon OS 5
        "/usr/libexec/sftp-server",
    ]
    sftp_server_path = None
    for path in sftp_server_paths:
        if os.path.exists(path):
            sftp_server_path = path
    if sftp_server_path is None:
        pytest.fail(f"Failed to find 'sftp-server'. Searched: {sftp_server_paths}")
    else:
        sshd_config_dict["Subsystem"] = f"sftp {sftp_server_path}"
    factory = salt_factories.get_sshd_daemon(
        sshd_config_dict=sshd_config_dict,
        config_dir=sshd_config_dir,
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="module")
def known_hosts_file(sshd_server, salt_master, salt_factories):
    with pytest.helpers.temp_file(
        "ssh-known-hosts",
        "\n".join(sshd_server.get_host_keys()),
        salt_factories.tmp_root_dir,
    ) as known_hosts_file, pytest.helpers.temp_file(
        "master.d/ssh-known-hosts.conf",
        f"known_hosts_file: {known_hosts_file}",
        salt_master.config_dir,
    ):
        yield known_hosts_file


@pytest.fixture(scope="module")
def salt_ssh_roster_file(sshd_server, salt_master, known_hosts_file):
    roster_contents = """
    localhost:
      host: 127.0.0.1
      port: {}
      user: {}
      mine_functions:
        test.arg: ['itworked']
    """.format(
        sshd_server.listen_port, RUNTIME_VARS.RUNNING_TESTS_USER
    )
    if salt.utils.platform.is_darwin():
        roster_contents += "  set_path: $PATH:/usr/local/bin/\n"

    with pytest.helpers.temp_file(
        "roster", roster_contents, salt_master.config_dir
    ) as roster_file:
        yield roster_file


# <---- Salt Factories -----------------------------------------------------------------------------------------------


# ----- From Filenames Test Selection ------------------------------------------------------------------------------->
def _match_to_test_file(match):
    parts = match.split(".")
    test_module_path = TESTS_DIR.joinpath(*parts)
    if test_module_path.exists():
        return test_module_path
    parts[-1] += ".py"
    return TESTS_DIR.joinpath(*parts).relative_to(CODE_DIR)


def from_filenames_collection_modifyitems(config, items):
    from_filenames = config.getoption("--from-filenames")
    if not from_filenames:
        # Don't do anything
        return

    terminal_reporter = config.pluginmanager.getplugin("terminalreporter")
    terminal_reporter.ensure_newline()
    terminal_reporter.section(
        "From Filenames(--from-filenames) Test Selection", sep=">"
    )
    errors = []
    test_module_selections = []
    changed_files_selections = []
    from_filenames_paths = set()
    for path in [path.strip() for path in from_filenames.split(",")]:
        # Make sure that, no matter what kind of path we're passed, Windows or Posix path,
        # we resolve it to the platform slash separator
        properly_slashed_path = pathlib.Path(
            path.replace("\\", os.sep).replace("/", os.sep)
        )
        if not properly_slashed_path.exists():
            errors.append(f"{properly_slashed_path}: Does not exist")
            continue
        if (
            properly_slashed_path.name == "testrun-changed-files.txt"
            or properly_slashed_path.is_absolute()
        ):
            # In this case, this path is considered to be a file containing a line separated list
            # of files to consider
            contents = properly_slashed_path.read_text(encoding="utf-8")
            for sep in ("\r\n", "\\r\\n", "\\n"):
                contents = contents.replace(sep, "\n")
            for line in contents.split("\n"):
                line_path = pathlib.Path(
                    line.strip().replace("\\", os.sep).replace("/", os.sep)
                )
                if not line_path.exists():
                    errors.append(
                        "{}: Does not exist. Source {}".format(
                            line_path, properly_slashed_path
                        )
                    )
                    continue
                changed_files_selections.append(
                    f"{line_path}: Source {properly_slashed_path}"
                )
                from_filenames_paths.add(line_path)
            continue
        changed_files_selections.append(
            f"{properly_slashed_path}: Source --from-filenames"
        )
        from_filenames_paths.add(properly_slashed_path)

    # Let's start collecting test modules
    test_module_paths = set()

    filename_map = yaml.deserialize((TESTS_DIR / "filename_map.yml").read_text())
    # Let's add the match all rule
    for rule, matches in filename_map.items():
        if rule == "*":
            for match in matches:
                test_module_paths.add(_match_to_test_file(match))
            break

    # Let's now go through the list of files gathered
    for path in from_filenames_paths:
        if path.as_posix().startswith("tests/"):
            if path.name == "conftest.py":
                # This is not a test module, but consider any test_*.py files in child directories
                for match in path.parent.rglob("test_*.py"):
                    test_module_selections.append(
                        "{}: Source '{}/test_*.py' recursive glob match because '{}' was modified".format(
                            match, path.parent, path
                        )
                    )
                    test_module_paths.add(match)
                continue
            # Tests in the listing don't require additional matching and will be added to the
            # list of tests to run
            test_module_selections.append(f"{path}: Source --from-filenames")
            test_module_paths.add(path)
            continue
        if path.name == "setup.py" or path.as_posix().startswith("salt/"):
            if path.name == "__init__.py":
                # No direct matching
                continue

            # Let's try a direct match between the passed file and possible test modules
            glob_patterns = (
                # salt/version.py ->
                #    tests/unit/test_version.py
                #    tests/pytests/unit/test_version.py
                f"**/test_{path.name}",
                # salt/modules/grains.py ->
                #    tests/pytests/integration/modules/grains/tests_*.py
                # salt/modules/saltutil.py ->
                #    tests/pytests/integration/modules/saltutil/test_*.py
                f"**/{path.stem}/test_*.py",
                # salt/modules/config.py ->
                #    tests/unit/modules/test_config.py
                #    tests/integration/modules/test_config.py
                #    tests/pytests/unit/modules/test_config.py
                #    tests/pytests/integration/modules/test_config.py
                f"**/{path.parent.name}/test_{path.name}",
            )
            for pattern in glob_patterns:
                for match in TESTS_DIR.rglob(pattern):
                    relative_path = match.relative_to(CODE_DIR)
                    test_module_selections.append(
                        "{}: Source '{}' glob pattern match".format(
                            relative_path, pattern
                        )
                    )
                    test_module_paths.add(relative_path)

            # Do we have an entry in tests/filename_map.yml
            for rule, matches in filename_map.items():
                if rule == "*":
                    continue
                elif "|" in rule:
                    # This is regex
                    if re.match(rule, path.as_posix()):
                        for match in matches:
                            test_module_paths.add(_match_to_test_file(match))
                            test_module_selections.append(
                                "{}: Source '{}' regex match from 'tests/filename_map.yml'".format(
                                    match, rule
                                )
                            )
                elif "*" in rule or "\\" in rule:
                    # Glob matching
                    for filerule in CODE_DIR.glob(rule):
                        if not filerule.exists():
                            continue
                        filerule = filerule.relative_to(CODE_DIR)
                        if filerule != path:
                            continue
                        for match in matches:
                            match_path = _match_to_test_file(match)
                            test_module_selections.append(
                                "{}: Source '{}' file rule from 'tests/filename_map.yml'".format(
                                    match_path, filerule
                                )
                            )
                            test_module_paths.add(match_path)
                else:
                    if path.as_posix() != rule:
                        continue
                    # Direct file paths as rules
                    filerule = pathlib.Path(rule)
                    if not filerule.exists():
                        continue
                    for match in matches:
                        match_path = _match_to_test_file(match)
                        test_module_selections.append(
                            "{}: Source '{}' direct file rule from 'tests/filename_map.yml'".format(
                                match_path, filerule
                            )
                        )
                        test_module_paths.add(match_path)
            continue
        else:
            errors.append(f"{path}: Don't know what to do with this path")

    if errors:
        terminal_reporter.write("Errors:\n", bold=True)
        for error in errors:
            terminal_reporter.write(f" * {error}\n")
    if changed_files_selections:
        terminal_reporter.write("Changed files collected:\n", bold=True)
        for selection in changed_files_selections:
            terminal_reporter.write(f" * {selection}\n")
    if test_module_selections:
        terminal_reporter.write("Selected test modules:\n", bold=True)
        for selection in test_module_selections:
            terminal_reporter.write(f" * {selection}\n")
    terminal_reporter.section(
        "From Filenames(--from-filenames) Test Selection", sep="<"
    )
    terminal_reporter.ensure_newline()

    selected = []
    deselected = []
    for item in items:
        itempath = pathlib.Path(str(item.fspath)).resolve().relative_to(CODE_DIR)
        if itempath in test_module_paths:
            selected.append(item)
        else:
            deselected.append(item)

    items[:] = selected
    if deselected:
        config.hook.pytest_deselected(items=deselected)


# <---- From Filenames Test Selection --------------------------------------------------------------------------------


# ----- Custom Fixtures --------------------------------------------------------------------------------------------->
@pytest.fixture(scope="session")
def sminion():
    return create_sminion()


@pytest.fixture(scope="session")
def grains(sminion):
    return sminion.opts["grains"].copy()


@lru_cache(maxsize=1)
def _grains_for_marker():
    return create_sminion().opts["grains"]


@pytest.fixture(scope="session", autouse=True)
def _disable_salt_logging():
    # This fixture is used to set logging to a configuration that salt expects,
    # however, no logging is actually configured since pytest's logging will be
    # logging what we need.
    logging_config = {
        # Undocumented, on purpose, at least for now, options.
        "configure_ext_handlers": False,
        "configure_file_logger": False,
        "configure_console_logger": False,
        "configure_granular_levels": False,
    }
    salt._logging.set_logging_options_dict(logging_config)
    # Run the test suite
    yield


@pytest.fixture(scope="session")
def salt_api_account_factory():
    return TestAccount(username="saltdev_api", password="saltdev")


# <---- Custom Fixtures ----------------------------------------------------------------------------------------------



import salt.loader.context
import salt.loader.lazy
import salt.utils.files
import tests.support.helpers


def test_opts_dunder_opts_without_import(tmp_path):
    """
    Test __opts__ without being imported.

    When a loaded module uses __opts__ but does not import it from
    salt.loader.dunder the __opts__ object will be a dictionary.
    """
    opts = {"optimization_order": [0, 1, 2]}
    with salt.utils.files.fopen(tmp_path / "mymod.py", "w") as fp:
        fp.write(
            tests.support.helpers.dedent(
                """
            def mymethod():
                return __opts__
            """
            )
        )
    loader = salt.loader.lazy.LazyLoader([tmp_path], opts)
    assert type(loader["mymod.mymethod"]()) == dict


def test_opts_dunder_opts_with_import(tmp_path):
    """
    Test __opts__ when imported.

    When a loaded module uses __opts__ by importing it from
    salt.loader.dunder the __opts__ object will be a NamedLoaderContext.
    """
    opts = {"optimization_order": [0, 1, 2]}
    with salt.utils.files.fopen(tmp_path / "mymod.py", "w") as fp:
        fp.write(
            tests.support.helpers.dedent(
                """
            from salt.loader.dunder import __opts__
            def optstype():
                return type(__opts__)
            def opts():
                return __opts__
            """
            )
        )
    loader = salt.loader.lazy.LazyLoader([tmp_path], opts)
    assert loader["mymod.optstype"]() == salt.loader.context.NamedLoaderContext
    assert loader["mymod.opts"]() == opts



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.svn
"""

import pytest

import salt.modules.svn as svn
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {svn: {}}


def test_info():
    """
    Test to display the Subversion information from the checkout.
    """
    mock = MagicMock(
        side_effect=[
            {"retcode": 0, "stdout": True},
            {"retcode": 0, "stdout": "A\n\nB"},
            {"retcode": 0, "stdout": "A\n\nB"},
        ]
    )
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.info("cwd", fmt="xml")

        assert svn.info("cwd", fmt="list") == [[], []]

        assert svn.info("cwd", fmt="dict") == [{}, {}]


def test_checkout():
    """
    Test to download a working copy of the remote Subversion repository
    directory or file
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.checkout("cwd", "remote")


def test_switch():
    """
    Test to switch a working copy of a remote Subversion repository
    directory
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.switch("cwd", "remote")


def test_update():
    """
    Test to update the current directory, files, or directories from
    the remote Subversion repository
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.update("cwd")


def test_diff():
    """
    Test to return the diff of the current directory, files, or
    directories from the remote Subversion repository
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.diff("cwd")


def test_commit():
    """
    Test to commit the current directory, files, or directories to
    the remote Subversion repository
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.commit("cwd")


def test_add():
    """
    Test to add files to be tracked by the Subversion working-copy
    checkout
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.add("cwd", False)


def test_remove():
    """
    Test to remove files and directories from the Subversion repository
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.remove("cwd", False)


def test_status():
    """
    Test to display the status of the current directory, files, or
    directories in the Subversion repository
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.status("cwd")


def test_export():
    """
    Test to create an unversioned copy of a tree.
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": True})
    with patch.dict(svn.__salt__, {"cmd.run_all": mock}):
        assert svn.export("cwd", "remote")



"""
    :codeauthor: Megan Wilhite<mwilhite@saltstack.com>
"""

import pytest

import salt.modules.mac_service as mac_service
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {mac_service: {"__context__": {}}}


def test_service_disabled_when_enabled():
    """
    test service.disabled when service is enabled
    """
    srv_name = "com.apple.atrun"
    cmd = (
        'disabled services = {\n\t"com.saltstack.salt.minion" =>'
        ' false\n\t"com.apple.atrun" => false\n{'
    )
    domain_ret = MagicMock(return_value=("", ""))
    with patch.object(mac_service, "_get_domain_target", domain_ret):
        with patch.object(mac_service, "launchctl", MagicMock(return_value=cmd)):
            assert mac_service.disabled(srv_name) is False


def test_service_disabled_when_disabled():
    """
    test service.disabled when service is disabled
    """
    srv_name = "com.apple.atrun"
    cmd = (
        'disabled services = {\n\t"com.saltstack.salt.minion" =>'
        ' false\n\t"com.apple.atrun" => true\n{'
    )
    domain_ret = MagicMock(return_value=("", ""))
    with patch.object(mac_service, "_get_domain_target", domain_ret):
        with patch.object(mac_service, "launchctl", MagicMock(return_value=cmd)):
            assert mac_service.disabled(srv_name) is True


def test_service_disabled_srvname_wrong():
    """
    test service.disabled when service is just slightly wrong
    """
    srv_names = ["com.apple.atru", "com", "apple"]
    cmd = (
        'disabled services = {\n\t"com.saltstack.salt.minion" =>'
        ' false\n\t"com.apple.atrun" => true\n}'
    )
    domain_ret = MagicMock(return_value=("", ""))
    with patch.object(mac_service, "_get_domain_target", domain_ret):
        for name in srv_names:
            with patch.object(mac_service, "launchctl", MagicMock(return_value=cmd)):
                assert mac_service.disabled(name) is False


def test_service_disabled_status_upper_case():
    """
    test service.disabled when disabled status is uppercase
    """
    srv_name = "com.apple.atrun"
    cmd = (
        'disabled services = {\n\t"com.saltstack.salt.minion" =>'
        ' false\n\t"com.apple.atrun" => True\n{'
    )
    domain_ret = MagicMock(return_value=("", ""))
    with patch.object(mac_service, "_get_domain_target", domain_ret):
        with patch.object(mac_service, "launchctl", MagicMock(return_value=cmd)):
            assert mac_service.disabled(srv_name) is True


def test_service_enabled_when_enabled():
    """
    test service.enabled when not disabled
    """
    mock_cmd = MagicMock(return_value=False)
    with patch.dict(mac_service.__salt__, {"service.disabled": mock_cmd}):
        assert mac_service.enabled("com.apple.atrun") is True


def test_service_enabled_when_disabled():
    """
    test service.enabled if service is disabled
    """
    mock_cmd = MagicMock(return_value=True)
    with patch.dict(mac_service.__salt__, {"service.disabled": mock_cmd}):
        assert mac_service.enabled("com.apple.atrun") is False


def test_service_loaded_when_true():
    """
    test service.loaded with a loaded service.
    """
    mock_cmd = MagicMock(return_value="some_service_string")
    with patch.dict(mac_service.__salt__, {"service.list": mock_cmd}):
        assert mac_service.loaded("com.apple.atrun") is True


def test_service_loaded_when_false():
    """
    test service.loaded with an unloaded service.
    """
    mock_cmd = MagicMock(side_effect=CommandExecutionError)
    with patch.dict(mac_service.__salt__, {"service.list": mock_cmd}):
        assert mac_service.loaded("com.apple.atrun") is False


def test_service_keep_alive_pathstate_file_rm():
    """
    test _always_running_service when keep_alive
    has pathstate set in plist file and file doesn't exist
    """
    srv_name = "com.apple.atrun"
    info = {
        "plist": {
            "EnableTransactions": True,
            "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
            "Label": "org.ntp.ntpd",
            "KeepAlive": {"PathState": {"/private/etc/ntp.conf": True}},
        }
    }

    with patch.object(mac_service, "show", MagicMock(return_value=info)):
        with patch("os.path.exists", MagicMock(return_value=False)):
            assert mac_service._always_running_service(srv_name) is False


def test_service_keep_alive_empty():
    """
    test _always_running_service when keep_alive
    is empty
    """
    srv_name = "com.apple.atrun"
    info = {
        "plist": {
            "EnableTransactions": True,
            "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
            "Label": "org.ntp.ntpd",
            "KeepAlive": {},
        }
    }

    with patch.object(mac_service, "show", MagicMock(return_value=info)):
        with patch("os.path.exists", MagicMock(return_value=False)):
            assert mac_service._always_running_service(srv_name) is False


def test_service_keep_alive_pathstate_false():
    """
    test _always_running_service when keep_alive
    has pathstate set in plist file and file is false
    """
    srv_name = "com.apple.atrun"
    info = {
        "plist": {
            "EnableTransactions": True,
            "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
            "Label": "org.ntp.ntpd",
            "KeepAlive": {"PathState": {"/private/etc/ntp.conf": False}},
        }
    }

    with patch.object(mac_service, "show", MagicMock(return_value=info)):
        with patch("os.path.exists", MagicMock(return_value=False)):
            assert mac_service._always_running_service(srv_name) is True


def test_service_keep_alive_pathstate():
    """
    test _always_running_service when keep_alive
    has pathstate set in plist file
    """
    srv_name = "com.apple.atrun"
    info = {
        "plist": {
            "EnableTransactions": True,
            "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
            "Label": "org.ntp.ntpd",
            "KeepAlive": {"PathState": {"/private/etc/ntp.conf": True}},
        }
    }

    with patch.object(mac_service, "show", MagicMock(return_value=info)):
        with patch("os.path.exists", MagicMock(return_value=True)):
            assert mac_service._always_running_service(srv_name) is True


def test_service_keep_alive():
    """
    test _always_running_service when keep_alive set
    """
    srv_name = "com.apple.atrun"
    info = {
        "plist": {
            "EnableTransactions": True,
            "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
            "Label": "org.ntp.ntpd",
            "KeepAlive": True,
        }
    }

    with patch.object(mac_service, "show", MagicMock(return_value=info)):
        assert mac_service._always_running_service(srv_name) is True


def test_service_keep_alive_false():
    """
    test _always_running_service when keep_alive False
    """
    srv_name = "com.apple.atrun"
    info = {
        "plist": {
            "EnableTransactions": True,
            "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
            "Label": "org.ntp.ntpd",
            "KeepAlive": False,
        }
    }

    with patch.object(mac_service, "show", MagicMock(return_value=info)):
        assert mac_service._always_running_service(srv_name) is False


def test_service_keep_alive_missing():
    """
    test _always_running_service when keep_alive not in dict
    """
    srv_name = "com.apple.atrun"
    info = {
        "plist": {
            "EnableTransactions": True,
            "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
            "Label": "org.ntp.ntpd",
        }
    }

    with patch.object(mac_service, "show", MagicMock(return_value=info)):
        assert mac_service._always_running_service(srv_name) is False


def test_service_keep_alive_wrong_setting():
    """
    test _always_running_service when keep_alive
    has pathstate set in plist file
    """
    srv_name = "com.apple.atrun"
    info = {
        "plist": {
            "EnableTransactions": True,
            "ProgramArguments": ["/usr/libexec/ntpd-wrapper"],
            "Label": "org.ntp.ntpd",
            "KeepAlive": {"Doesnotexist": {"doesnt_exist": True}},
        }
    }

    with patch.object(mac_service, "show", MagicMock(return_value=info)):
        assert mac_service._always_running_service(srv_name) is False


def test_service_name_change_salt_minion():
    srv_name = "salt-minion"
    info = {
        "com.saltstack.salt.minion": {
            "file_name": "com.saltstack.salt.minion.plist",
            "file_path": "/Library/LaunchDaemons/com.saltstack.salt.minion.plist",
            "plist": {
                "HardResourceLimits": {"NumberOfFiles": 100000},
                "KeepAlive": True,
                "Label": "com.saltstack.salt.minion",
                "ProgramArguments": ["/opt/salt/bin/start-salt-minion.sh"],
                "RunAtLoad": True,
                "SoftResourceLimits": {"NumberOfFiles": 100000},
            },
        }
    }
    with patch.dict(
        mac_service.__utils__,
        {"mac_utils.available_services": MagicMock(return_value=info)},
    ):
        assert mac_service._get_service(srv_name) == info["com.saltstack.salt.minion"]


def test_service_name_change_salt_master():
    srv_name = "salt-master"
    info = {
        "com.saltstack.salt.master": {
            "file_name": "com.saltstack.salt.master.plist",
            "file_path": "/Library/LaunchDaemons/com.saltstack.salt.master.plist",
            "plist": {
                "HardResourceLimits": {"NumberOfFiles": 100000},
                "KeepAlive": True,
                "Label": "com.saltstack.salt.master",
                "ProgramArguments": ["/opt/salt/bin/start-salt-master.sh"],
                "RunAtLoad": True,
                "SoftResourceLimits": {"NumberOfFiles": 100000},
            },
        }
    }
    with patch.dict(
        mac_service.__utils__,
        {"mac_utils.available_services": MagicMock(return_value=info)},
    ):
        assert mac_service._get_service(srv_name) == info["com.saltstack.salt.master"]


def test_service_name_change_salt_api():
    srv_name = "salt-api"
    info = {
        "com.saltstack.salt.api": {
            "file_name": "com.saltstack.salt.api.plist",
            "file_path": "/Library/LaunchDaemons/com.saltstack.salt.api.plist",
            "plist": {
                "HardResourceLimits": {"NumberOfFiles": 100000},
                "KeepAlive": True,
                "Label": "com.saltstack.salt.api",
                "ProgramArguments": ["/opt/salt/bin/start-salt-api.sh"],
                "RunAtLoad": True,
                "SoftResourceLimits": {"NumberOfFiles": 100000},
            },
        }
    }
    with patch.dict(
        mac_service.__utils__,
        {"mac_utils.available_services": MagicMock(return_value=info)},
    ):
        assert mac_service._get_service(srv_name) == info["com.saltstack.salt.api"]


def test_service_name_change_salt_syndic():
    srv_name = "salt-syndic"
    info = {
        "com.saltstack.salt.syndic": {
            "file_name": "com.saltstack.salt.syndic.plist",
            "file_path": "/Library/LaunchDaemons/com.saltstack.salt.syndic.plist",
            "plist": {
                "HardResourceLimits": {"NumberOfFiles": 100000},
                "KeepAlive": True,
                "Label": "com.saltstack.salt.syndic",
                "ProgramArguments": ["/opt/salt/bin/start-salt-syndic.sh"],
                "RunAtLoad": True,
                "SoftResourceLimits": {"NumberOfFiles": 100000},
            },
        }
    }
    with patch.dict(
        mac_service.__utils__,
        {"mac_utils.available_services": MagicMock(return_value=info)},
    ):
        assert mac_service._get_service(srv_name) == info["com.saltstack.salt.syndic"]


def test_service_restart_already_loaded():
    mock_cmd = MagicMock(return_value=True)
    salt_dict = {
        "service.loaded": mock_cmd,
        "service.stop": mock_cmd,
        "service.start": mock_cmd,
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        assert mac_service.restart("com.salt") is True


def test_service_restart_not_loaded():
    salt_dict = {
        "service.loaded": MagicMock(return_value=False),
        "service.start": MagicMock(return_value=True),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        assert mac_service.restart("com.salt") is True


def test_service_restart_failed_stop():
    salt_dict = {
        "service.loaded": MagicMock(return_value=True),
        "service.stop": MagicMock(side_effect=CommandExecutionError),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        with pytest.raises(CommandExecutionError):
            assert mac_service.restart("com.salt")


def test_service_restart_failed_start():
    salt_dict = {
        "service.loaded": MagicMock(return_value=False),
        "service.start": MagicMock(side_effect=CommandExecutionError),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        with pytest.raises(CommandExecutionError):
            assert mac_service.restart("com.salt")


def test_service_status_no_service():
    """
    Test service status with no service found
    """
    with patch.object(
        mac_service, "_get_service", MagicMock(side_effect=CommandExecutionError)
    ):
        assert mac_service.status("com.salt") is False


def test_service_status_on_daemon_with_pid():
    """
    Test service status on dameon with PID.
    """
    mock_service_list = (
        '{\n\t"LimitLoadToSessionType" = "System";\n\t"Label" ='
        ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" = 0;\n\t"PID" ='
        ' 218;\n\t"Program" = "/opt/salt";\n\t\t"--disable-keepalive";\n\t);\n};'
    )
    salt_dict = {
        "service.list": MagicMock(return_value=mock_service_list),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        with patch.object(mac_service, "_launch_agent", lambda _: False):
            with patch.object(mac_service, "_get_service", lambda _: {"": ""}):
                with patch.object(
                    mac_service, "_always_running_service", lambda _: True
                ):
                    assert mac_service.status("com.salt") is True


def test_service_status_on_agent_with_pid():
    """
    Test service status on LaunchAgent with PID.
    """
    mock_service_list = (
        '{\n\t"LimitLoadToSessionType" = "Aqua";\n\t"Label" ='
        ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" = 19968;\n\t"PID"'
        ' = 218;\n\t"Program" = "/opt/salt";\n\t"ProgramArguments" ='
        ' (\n\t\t"/opt/salt";\n\t\t"--syslog";\n\t);\n};'
    )
    salt_dict = {
        "service.list": MagicMock(return_value=mock_service_list),
    }
    utils_dict = {
        "mac_utils.console_user": MagicMock(return_value="spongebob"),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        with patch.dict(mac_service.__utils__, utils_dict):
            with patch.object(mac_service, "_launch_agent", lambda _: True):
                with patch.object(mac_service, "_get_service", lambda _: {"": ""}):
                    with patch.object(
                        mac_service, "_always_running_service", lambda _: True
                    ):
                        assert mac_service.status("com.salt") is True


def test_service_status_on_agent_with_no_pid_and_should_be_running():
    """
    Test service status on LaunchAgent with No PID and should be running.
    """
    mock_service_list = (
        '{\n\t"LimitLoadToSessionType" = "Aqua";\n\t"Label" ='
        ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" ='
        ' 19968;\n\t"Program" = "/opt/salt";\n\t"ProgramArguments" ='
        ' (\n\t\t"/opt/salt";\n\t\t"--syslog";\n\t);\n};'
    )
    salt_dict = {
        "service.list": MagicMock(return_value=mock_service_list),
    }
    utils_dict = {
        "mac_utils.console_user": MagicMock(return_value="spongebob"),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        with patch.dict(mac_service.__utils__, utils_dict):
            with patch.object(mac_service, "_launch_agent", lambda _: True):
                with patch.object(mac_service, "_get_service", lambda _: {"": ""}):
                    with patch.object(
                        mac_service, "_always_running_service", lambda _: True
                    ):
                        assert mac_service.status("com.salt") is False


def test_service_status_on_daemon_with_no_pid_and_should_be_running():
    """
    Test service status on LaunchDaemon with no PID and an
    always running service that is loaded.
    """
    mock_service_list = (
        '{\n\t"LimitLoadToSessionType" = "System";\n\t"Label" ='
        ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" ='
        ' 19968;\n\t"Program" = "/opt/salt.sh";\n\t"ProgramArguments" ='
        ' (\n\t\t"/opt/salt.sh";\n\t\t"--disable-keepalive";\n\t);\n};'
    )
    salt_dict = {
        "service.list": MagicMock(return_value=mock_service_list),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        with patch.object(mac_service, "_launch_agent", lambda _: False):
            with patch.object(mac_service, "_get_service", lambda _: {"": ""}):
                with patch.object(
                    mac_service, "_always_running_service", lambda _: True
                ):
                    assert mac_service.status("com.salt") is False


def test_service_status_on_daemon_with_no_pid_and_not_always_running():
    """
    Test service status on LaunchDaemon with no PID and not an always
    running service.
    """
    mock_service_list = (
        '{\n\t"LimitLoadToSessionType" = "System";\n\t"Label" ='
        ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" ='
        ' 19968;\n\t"Program" = "/opt/salt.sh";\n\t"ProgramArguments" ='
        ' (\n\t\t"/opt/salt.sh";\n\t\t"--disable-keepalive";\n\t);\n};'
    )
    salt_dict = {
        "service.list": MagicMock(return_value=mock_service_list),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        with patch.object(mac_service, "_launch_agent", lambda _: False):
            with patch.object(mac_service, "_get_service", lambda _: {"": ""}):
                with patch.object(
                    mac_service, "_always_running_service", lambda _: False
                ):
                    assert mac_service.status("com.salt") is True


def test_service_status_on_daemon_with_failing_list_check():
    """
    Test service status on LaunchDaemon with no PID on an
    always running service that is loaded.
    """
    mock_service_list = (
        '{\n\t"LimitLoadToSessionType" = "System";\n\t"Label" ='
        ' "com.salt";\n\t"OnDemand" = false;\n\t"LastExitStatus" ='
        ' 19968;\n\t"Program" = "/opt/salt.sh";\n\t"ProgramArguments" ='
        ' (\n\t\t"/opt/salt.sh";\n\t\t"--disable-keepalive";\n\t);\n};'
    )
    salt_dict = {
        "service.list": MagicMock(side_effect=CommandExecutionError),
    }
    with patch.dict(mac_service.__salt__, salt_dict):
        with patch.object(mac_service, "_launch_agent", lambda _: False):
            with patch.object(mac_service, "_get_service", lambda _: {"": ""}):
                with patch.object(
                    mac_service, "_always_running_service", lambda _: False
                ):
                    assert mac_service.status("com.salt") is False


def test_get_service_on_service_dead():
    """
    Test service.dead changes.
    https://github.com/saltstack/salt/issues/57907
    """
    utils_dict = {
        "mac_utils.available_services": MagicMock(return_value={}),
    }
    context_dict = {
        "using_cached_services": True,
        "service.state": "dead",
    }
    name_in_service = MagicMock(side_effect=[{}, {"com.salt": True}])
    with patch.dict(mac_service.__utils__, utils_dict):
        with patch.object(mac_service, "_name_in_services", name_in_service):
            with patch.dict(mac_service.__context__, context_dict):
                with pytest.raises(CommandExecutionError):
                    assert mac_service._get_service("com.salt")
            # find the service on a second go with no service.dead
            with patch.dict(mac_service.__context__, {}):
                assert mac_service._get_service("com.salt") == {"com.salt": True}



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import os

import pytest
from saltfactories.utils import random_string

import salt.utils.files

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def user(modules):
    return modules.user


@pytest.fixture
def _reset_enable_auto_login(user):
    # Make sure auto login is disabled before we start
    if user.get_auto_login():
        pytest.skip("Auto login already enabled")

    try:
        yield
    finally:
        # Make sure auto_login is disabled
        ret = user.disable_auto_login()
        assert ret

        # Make sure autologin is disabled
        ret = user.get_auto_login()
        if ret:
            pytest.fail("Failed to disable auto login")


@pytest.fixture
def existing_user(user):
    username = random_string("account-", uppercase=False)
    try:
        ret = user.add(username)
        if ret is not True:
            pytest.skip("Failed to create an account to manipulate")
        yield username
    finally:
        user_info = user.info(username)
        if user_info:
            user.delete(username)


@pytest.fixture
def non_existing_user(user):
    username = random_string("account-", uppercase=False)
    try:
        yield username
    finally:
        user_info = user.info(username)
        if user_info:
            user.delete(username)


def test_mac_user_add(user, non_existing_user):
    """
    Tests the add function
    """
    user.add(non_existing_user)
    user_info = user.info(non_existing_user)
    assert user_info["name"] == non_existing_user


def test_mac_user_delete(user, existing_user):
    """
    Tests the delete function
    """
    ret = user.delete(existing_user)
    assert ret


def test_mac_user_primary_group(user, existing_user):
    """
    Tests the primary_group function
    """
    primary_group = user.primary_group(existing_user)
    uid_info = user.info(existing_user)
    assert primary_group in uid_info["groups"]


def test_mac_user_changes(user, existing_user):
    """
    Tests mac_user functions that change user properties
    """
    # Test mac_user.chuid
    user.chuid(existing_user, 4376)
    uid_info = user.info(existing_user)
    assert uid_info["uid"] == 4376

    # Test mac_user.chgid
    user.chgid(existing_user, 4376)
    gid_info = user.info(existing_user)
    assert gid_info["gid"] == 4376

    # Test mac.user.chshell
    user.chshell(existing_user, "/bin/zsh")
    shell_info = user.info(existing_user)
    assert shell_info["shell"] == "/bin/zsh"

    # Test mac_user.chhome
    user.chhome(existing_user, "/Users/foo")
    home_info = user.info(existing_user)
    assert home_info["home"] == "/Users/foo"

    # Test mac_user.chfullname
    user.chfullname(existing_user, "Foo Bar")
    fullname_info = user.info(existing_user)
    assert fullname_info["fullname"] == "Foo Bar"

    # Test mac_user.chgroups
    ret = user.info(existing_user)
    pre_info = ret["groups"]
    expected = pre_info + ["wheel"]
    user.chgroups(existing_user, "wheel")
    groups_info = user.info(existing_user)
    assert groups_info["groups"] == expected


@pytest.mark.usefixtures("_reset_enable_auto_login")
def test_mac_user_enable_auto_login(user):
    """
    Tests mac_user functions that enable auto login
    """
    # Does enable return True
    ret = user.enable_auto_login("Spongebob", "Squarepants")
    assert ret

    # Did it set the user entry in the plist file
    ret = user.get_auto_login()
    assert ret == "Spongebob"

    # Did it generate the `/etc/kcpassword` file
    assert os.path.exists("/etc/kcpassword")

    # Are the contents of the file correct
    test_data = bytes.fromhex("2e f8 27 42 a0 d9 ad 8b cd cd 6c 7d")
    with salt.utils.files.fopen("/etc/kcpassword", "rb") as f:
        file_data = f.read()
    assert test_data == file_data

    # Does disable return True
    ret = user.disable_auto_login()
    assert ret

    # Does it remove the user entry in the plist file
    ret = user.get_auto_login()
    assert not ret

    # Is the `/etc/kcpassword` file removed
    assert not os.path.exists("/etc/kcpassword")


@pytest.mark.usefixtures("_reset_enable_auto_login")
def test_mac_user_disable_auto_login(user):
    """
    Tests mac_user functions that disable auto login
    """
    # Enable auto login for the test
    user.enable_auto_login("Spongebob", "Squarepants")

    # Make sure auto login got set up
    ret = user.get_auto_login()
    if not ret == "Spongebob":
        raise pytest.fail("Failed to enable auto login")

    # Does disable return True
    ret = user.disable_auto_login()
    assert ret

    # Does it remove the user entry in the plist file
    ret = user.get_auto_login()
    assert not ret

    # Is the `/etc/kcpassword` file removed
    assert not os.path.exists("/etc/kcpassword")



import shutil
import tempfile
from pathlib import Path

import psutil  # pylint: disable=3rd-party-module-not-gated
import pytest
from pytestshellutils.utils.processes import terminate_process

import salt.fileserver.hgfs as hgfs
from tests.support.mock import patch

try:
    import hglib

    HAS_HG = True
except ImportError:
    HAS_HG = False


@pytest.fixture(scope="module")
def configure_loader_modules(master_opts):
    master_opts["fileserver_backend"] = ["hgfs"]
    yield {hgfs: {"__opts__": master_opts}}


@pytest.fixture
def hgfs_setup_and_teardown():
    """
    build up and tear down hg repos to test with.
    """
    initial_child_processes = psutil.Process().children()
    source_dir = Path(__file__).resolve().parent.joinpath("files")
    tempdir = tempfile.TemporaryDirectory()
    tempsubdir = tempdir.name / Path("subdir/")
    tempsubdir.mkdir()
    tempdirPath = Path(tempdir.name)
    for file in source_dir.iterdir():
        to_file = tempdirPath / file.name
        to_file2 = tempsubdir / file.name
        shutil.copy(file.as_posix(), to_file.as_posix())
        shutil.copy(file.as_posix(), to_file2.as_posix())

    client = hglib.init(bytes(tempdirPath.as_posix(), encoding="utf8"))
    client.close()
    with hglib.open(bytes(tempdirPath.as_posix(), encoding="utf8")) as repo:
        repo.add(bytes(tempdirPath.as_posix(), encoding="utf8"))
        repo.commit(b"init commit", user="test")
        repo.tag(b"test", user="test")
        repo.branch(b"test")
        repo.commit(b"create test branch", user="test")
        repo.bookmark(b"bookmark_test")
    try:
        yield tempdirPath.as_uri()
    finally:
        tempdir.cleanup()
        for child in psutil.Process().children():
            if child not in initial_child_processes:
                terminate_process(process=child, kill_children=True)


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_fix_58852(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repo = hgfs.init()
        assert isinstance(repo, list)
        if isinstance(repo, list):
            for value in repo:
                assert isinstance(value, dict)
                for key, value in value.items():
                    if key != "repo":
                        assert isinstance(value, str)


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_all_branches(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repos = hgfs.init()
        hgfs.update()
        for repo in repos:
            repo["repo"].open()
            branches = hgfs._all_branches(repo["repo"])
            assert isinstance(branches, list)
            if isinstance(branches, list):
                for value in branches:
                    assert isinstance(value, tuple)
                    assert len(value) == 3
                    assert value[0] in ["default", "test"]
                    assert isinstance(value[1], int)
                    assert isinstance(value[2], str)


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_get_branch(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repo = hgfs.init()
        hgfs.update()
        repo[0]["repo"].open()
        branch = hgfs._get_branch(repo[0]["repo"], "test")
        assert isinstance(branch, tuple)
        assert len(branch) == 3
        assert branch[0] in "test"
        assert branch[1] == 2
        assert isinstance(branch[2], str)

        # Fail test
        branch = hgfs._get_branch(repo[0]["repo"], "fake")
        assert branch is False


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_all_bookmarks(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repos = hgfs.init()
        hgfs.update()
        for repo in repos:
            repo["repo"].open()
            bookmarks = hgfs._all_bookmarks(repo["repo"])
            assert isinstance(bookmarks, list)
            if isinstance(bookmarks, list):
                for value in bookmarks:
                    assert isinstance(value, tuple)
                    assert len(value) == 3
                    assert value[0] in ["bookmark_test"]
                    assert value[1] == 2
                    assert isinstance(value[2], str)


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_get_bookmark(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repo = hgfs.init()
        hgfs.update()
        repo[0]["repo"].open()
        bookmark = hgfs._get_bookmark(repo[0]["repo"], "bookmark_test")
        assert isinstance(bookmark, tuple)
        assert len(bookmark) == 3
        assert bookmark[0] in "bookmark_test"
        assert bookmark[1] == 2
        assert isinstance(bookmark[2], str)

        # Fail test
        bookmark = hgfs._get_bookmark(repo[0]["repo"], "fake")
        assert bookmark is False


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_all_tags(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repos = hgfs.init()
        hgfs.update()
        for repo in repos:
            repo["repo"].open()
            tags = hgfs._all_tags(repo["repo"])
            assert isinstance(tags, list)
            if isinstance(tags, list):
                for value in tags:
                    assert isinstance(value, tuple)
                    assert len(value) == 4
                    assert value[0] in ["test"]
                    assert value[0] not in ["tip"]
                    assert value[1] == 0
                    assert isinstance(value[2], str)
                    assert value[3] is False


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_get_tag(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repo = hgfs.init()
        hgfs.update()
        repo[0]["repo"].open()
        tag = hgfs._get_tag(repo[0]["repo"], "test")
        assert isinstance(tag, tuple)
        assert len(tag) == 4
        assert tag[0] in "test"
        assert tag[1] == 0
        assert isinstance(tag[2], str)

        # Fail test
        tag = hgfs._get_tag(repo[0]["repo"], "fake")
        assert tag is False

        # real tag that should fail
        tag = hgfs._get_tag(repo[0]["repo"], "tip")
        assert tag is False


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_get_ref(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repo = hgfs.init()[0]
        hgfs.update()
        repo["repo"].open()
        ref = hgfs._get_ref(repo, "base")
        assert isinstance(ref, tuple)
        assert len(ref) == 3
        assert ref[0] == "default"
        assert ref[1] == 1
        assert isinstance(ref[2], str)

    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [
                {str(hgfs_setup_and_teardown): [{"base": "bookmark_test"}]}
            ],
            "hgfs_branch_method": "bookmarks",
        },
    ):
        repo = hgfs.init()[0]
        hgfs.update()
        repo["repo"].open()
        ref = hgfs._get_ref(repo, "base")
        assert isinstance(ref, tuple)
        assert len(ref) == 3
        assert ref[0] in "bookmark_test"
        assert ref[1] == 2
        assert isinstance(ref[2], str)

    # Fail test
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repo = hgfs.init()[0]
        hgfs.update()
        repo["repo"].open()
        ref = hgfs._get_ref(repo, "fake")
        assert ref is False


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_get_manifest(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        repo = hgfs.init()[0]
        hgfs.update()
        repo["repo"].open()
        ref = hgfs._get_ref(repo, "base")
        manifest = hgfs._get_manifest(repo["repo"], ref=ref)
        assert isinstance(manifest, list)
        for value in manifest:
            assert len(value) == 5
            assert isinstance(value[0], str)
            assert value[1] == "644"
            assert value[2] is False
            assert value[3] is False
            assert value[4] in [
                "test.sls",
                "test2.sls",
                ".hgtags",
                "subdir/test.sls",
                "subdir/test2.sls",
            ]


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_envs(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
            "hgfs_branch_method": "branches",
        },
    ):
        hgfs.init()
        hgfs.update()
        envs = hgfs.envs(ignore_cache=True)
        assert isinstance(envs, list)
        assert envs == ["base", "test"]
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [
                {str(hgfs_setup_and_teardown): [{"base": "bookmark_test"}]}
            ],
            "hgfs_branch_method": "bookmarks",
        },
    ):
        hgfs.init()
        hgfs.update()
        envs = hgfs.envs(ignore_cache=True)
        assert isinstance(envs, list)

        # apperently test is coming from the tags which will always be included in the envs unless blacklisted.
        # Do we really want that behavior?
        assert envs == ["base", "test"]


@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_env_is_exposed_blacklist():
    with patch.dict(
        hgfs.__opts__,
        {"hgfs_saltenv_whitelist": "", "hgfs_saltenv_blacklist": "test"},
    ):
        hgfs.init()
        hgfs.update()
        assert hgfs._env_is_exposed("base") is True
        assert hgfs._env_is_exposed("test") is False
        assert hgfs._env_is_exposed("unset") is True


@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_env_is_exposed_whitelist():
    with patch.dict(
        hgfs.__opts__,
        {"hgfs_saltenv_whitelist": "base", "hgfs_saltenv_blacklist": ""},
    ):
        hgfs.init()
        hgfs.update()
        assert hgfs._env_is_exposed("base") is True
        assert hgfs._env_is_exposed("test") is False
        assert hgfs._env_is_exposed("unset") is False


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_find_file(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        hgfs.init()
        hgfs.update()
        file = hgfs.find_file(path="test.sls", tgt_env="base")
        assert file["path"] == hgfs.__opts__["cachedir"] + "/hgfs/refs/base/test.sls"
        assert file["rel"] == "test.sls"
        assert isinstance(file["stat"], list)
        for i in file["stat"]:
            assert isinstance(i, int)


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_serve_file(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        hgfs.init()
        hgfs.update()
        file = hgfs.find_file(path="test.sls", tgt_env="base")
        load = {"saltenv": "base", "loc": 0, "path": "test.sls"}
        data = hgfs.serve_file(load, file)
        assert data == {
            "data": "always-passes:\n  test.succeed_without_changes:\n    - name: foo\n",
            "dest": "test.sls",
        }


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_file_hash(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        hgfs.init()
        hgfs.update()
        file = hgfs.find_file(path="test.sls", tgt_env="base")
        load = {"saltenv": "base", "loc": 0, "path": "test.sls"}
        data = hgfs.file_hash(load, file)
        assert data == {
            "hash_type": "sha256",
            "hsum": "a6a48d90dce9c9b580efb2ed308af100a8328913dcf9441705125866551c7d8d",
        }


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_file_list(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        hgfs.init()
        hgfs.update()
        load = {"saltenv": "base", "loc": 0, "path": "test.sls"}
        data = hgfs.file_list(load)
        assert data == [
            ".hgtags",
            "subdir/test.sls",
            "subdir/test2.sls",
            "test.sls",
            "test2.sls",
        ]


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(reason="testing break in windows")
def test_dir_list(hgfs_setup_and_teardown):
    with patch.dict(
        hgfs.__opts__,
        {
            "hgfs_remotes": [{str(hgfs_setup_and_teardown): [{"base": "default"}]}],
        },
    ):
        hgfs.init()
        hgfs.update()
        load = {"saltenv": "base", "loc": 0, "path": "test.sls"}
        data = hgfs.dir_list(load)
        assert data == ["subdir"]



"""
Manage the information in the hosts file
"""

import errno
import logging
import os
from collections import OrderedDict

import salt.utils.files
import salt.utils.stringutils

log = logging.getLogger(__name__)


# pylint: disable=C0103
def __get_hosts_filename():
    """
    Return the path to the appropriate hosts file
    """
    try:
        return __context__["hosts.__get_hosts_filename"]
    except KeyError:
        __context__["hosts.__get_hosts_filename"] = __salt__["config.option"](
            "hosts.file"
        )
        return __context__["hosts.__get_hosts_filename"]


def _get_or_create_hostfile():
    """
    Wrapper of __get_hosts_filename but create host file if it
    does not exist.
    """
    hfn = __get_hosts_filename()
    if hfn is None:
        hfn = ""
    if not os.path.exists(hfn):
        with salt.utils.files.fopen(hfn, "w"):
            pass
    return hfn


def _list_hosts():
    """
    Return the hosts found in the hosts file in as an OrderedDict
    """
    try:
        return __context__["hosts._list_hosts"]
    except KeyError:
        count = 0
        hfn = __get_hosts_filename()
        ret = OrderedDict()
        try:
            with salt.utils.files.fopen(hfn) as ifile:
                for line in ifile:
                    line = salt.utils.stringutils.to_unicode(line).strip()
                    if not line:
                        continue
                    if line.startswith("#"):
                        ret.setdefault(f"comment-{count}", []).append(line)
                        count += 1
                        continue
                    comment = None
                    if "#" in line:
                        comment = line[line.index("#") + 1 :].lstrip()
                        line = line[: line.index("#")].strip()
                    comps = line.split()
                    ip = comps.pop(0)
                    if comment:
                        ret.setdefault(ip, {}).setdefault("aliases", []).extend(comps)
                        ret.setdefault(ip, {}).update({"comment": comment})
                    else:
                        ret.setdefault(ip, {}).setdefault("aliases", []).extend(comps)
        except OSError as exc:
            salt.utils.files.process_read_exception(exc, hfn, ignore=errno.ENOENT)
            # Don't set __context__ since we weren't able to read from the
            # hosts file.
            return ret

        __context__["hosts._list_hosts"] = ret
        return ret


def list_hosts():
    """
    Return the hosts found in the hosts file in this format::

        {'<ip addr>': ['alias1', 'alias2', ...]}

    CLI Example:

    .. code-block:: bash

        salt '*' hosts.list_hosts
    """
    # msgpack does not like OrderedDict's
    return dict(_list_hosts())


def get_ip(host):
    """
    Return the ip associated with the named host

    CLI Example:

    .. code-block:: bash

        salt '*' hosts.get_ip <hostname>
    """
    hosts = _list_hosts()
    if not hosts:
        return ""
    # Look for the op
    for addr in hosts:
        if isinstance(hosts[addr], dict) and "aliases" in hosts[addr]:
            _hosts = hosts[addr]["aliases"]
            if host in _hosts:
                return addr
    # ip not found
    return ""


def get_alias(ip):
    """
    Return the list of aliases associated with an ip

    Aliases (host names) are returned in the order in which they
    appear in the hosts file.  If there are no aliases associated with
    the IP, an empty list is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' hosts.get_alias <ip addr>
    """
    hosts = _list_hosts()
    if ip in list(hosts):
        return hosts[ip]["aliases"]
    return []


def has_pair(ip, alias):
    """
    Return true if the alias is set

    CLI Example:

    .. code-block:: bash

        salt '*' hosts.has_pair <ip> <alias>
    """
    hosts = _list_hosts()
    try:
        if isinstance(alias, list):
            return set(alias).issubset(hosts[ip]["aliases"])
        else:
            return alias in hosts[ip]["aliases"]
    except KeyError:
        return False


def set_host(ip, alias, comment=None):
    """
    Set the host entry in the hosts file for the given ip, this will overwrite
    any previous entry for the given ip

    .. versionchanged:: 2016.3.0
        If ``alias`` does not include any host names (it is the empty
        string or contains only whitespace), all entries for the given
        IP address are removed.

    CLI Example:

    .. code-block:: bash

        salt '*' hosts.set_host <ip> <alias>
    """
    hfn = _get_or_create_hostfile()
    ovr = False
    if not os.path.isfile(hfn):
        return False

    # Make sure future calls to _list_hosts() will re-read the file
    __context__.pop("hosts._list_hosts", None)

    if comment:
        line_to_add = salt.utils.stringutils.to_bytes(
            ip + "\t\t" + alias + "\t\t# " + comment + os.linesep
        )
    else:
        line_to_add = salt.utils.stringutils.to_bytes(ip + "\t\t" + alias + os.linesep)
    # support removing a host entry by providing an empty string
    if not alias.strip():
        line_to_add = b""

    with salt.utils.files.fopen(hfn, "rb") as fp_:
        lines = fp_.readlines()
    for ind, _ in enumerate(lines):
        tmpline = lines[ind].strip()
        if not tmpline:
            continue
        if tmpline.startswith(b"#"):
            continue
        comps = tmpline.split()
        if comps[0] == salt.utils.stringutils.to_bytes(ip):
            if not ovr:
                lines[ind] = line_to_add
                ovr = True
            else:  # remove other entries
                lines[ind] = b""
    linesep_bytes = salt.utils.stringutils.to_bytes(os.linesep)
    if not ovr:
        # make sure there is a newline
        if lines and not lines[-1].endswith(linesep_bytes):
            lines[-1] += linesep_bytes
        line = line_to_add
        lines.append(line)
    with salt.utils.files.fopen(hfn, "wb") as ofile:
        ofile.writelines(lines)
    return True


def rm_host(ip, alias):
    """
    Remove a host entry from the hosts file

    CLI Example:

    .. code-block:: bash

        salt '*' hosts.rm_host <ip> <alias>
    """
    if not has_pair(ip, alias):
        return True
    # Make sure future calls to _list_hosts() will re-read the file
    __context__.pop("hosts._list_hosts", None)
    hfn = _get_or_create_hostfile()
    with salt.utils.files.fopen(hfn, "rb") as fp_:
        lines = fp_.readlines()
    for ind, _ in enumerate(lines):
        tmpline = lines[ind].strip()
        if not tmpline:
            continue
        if tmpline.startswith(b"#"):
            continue
        comps = tmpline.split()
        comment = None
        if b"#" in tmpline:
            host_info, comment = tmpline.split(b"#")
            comment = salt.utils.stringutils.to_bytes(comment).lstrip()
        else:
            host_info = tmpline
        host_info = salt.utils.stringutils.to_bytes(host_info)
        comps = host_info.split()
        b_ip = salt.utils.stringutils.to_bytes(ip)
        b_alias = salt.utils.stringutils.to_bytes(alias)
        if comps[0] == b_ip:
            newline = comps[0] + b"\t\t"
            for existing in comps[1:]:
                if existing == b_alias:
                    continue
                newline += existing + b" "
            if newline.strip() == b_ip:
                # No aliases exist for the line, make it empty
                lines[ind] = b""
            else:
                # Only an alias was removed
                if comment:
                    lines[ind] = (
                        newline
                        + b"# "
                        + comment
                        + salt.utils.stringutils.to_bytes(os.linesep)
                    )
                else:
                    lines[ind] = newline + salt.utils.stringutils.to_bytes(os.linesep)
    with salt.utils.files.fopen(hfn, "wb") as ofile:
        ofile.writelines(lines)
    return True


def add_host(ip, alias):
    """
    Add a host to an existing entry, if the entry is not in place then create
    it with the given host

    CLI Example:

    .. code-block:: bash

        salt '*' hosts.add_host <ip> <alias>
    """
    hfn = _get_or_create_hostfile()
    if not os.path.isfile(hfn):
        return False

    if has_pair(ip, alias):
        return True

    hosts = _list_hosts()

    # Make sure future calls to _list_hosts() will re-read the file
    __context__.pop("hosts._list_hosts", None)

    inserted = False
    for i, h in hosts.items():
        for num, host in enumerate(h):
            if isinstance(h, list):
                if host.startswith("#") and i == ip:
                    h.insert(num, alias)
                    inserted = True
    if not inserted:
        hosts.setdefault(ip, {}).setdefault("aliases", []).append(alias)
    _write_hosts(hosts)
    return True


def set_comment(ip, comment):
    """
    Set the comment for a host to an existing entry,
    if the entry is not in place then return False

    CLI Example:

    .. code-block:: bash

        salt '*' hosts.set_comment <ip> <comment>
    """
    hfn = _get_or_create_hostfile()
    if not os.path.isfile(hfn):
        return False

    hosts = _list_hosts()

    # Make sure future calls to _list_hosts() will re-read the file
    __context__.pop("hosts._list_hosts", None)

    if ip not in hosts:
        return False

    if "comment" in hosts[ip]:
        if comment != hosts[ip]["comment"]:
            hosts[ip]["comment"] = comment
            _write_hosts(hosts)
        else:
            return True
    else:
        hosts[ip]["comment"] = comment
        _write_hosts(hosts)
    return True


def _write_hosts(hosts):
    lines = []
    for ip, host_info in hosts.items():
        if ip:
            if ip.startswith("comment"):
                line = "".join(host_info)
            else:
                if "comment" in host_info:
                    line = "{}\t\t{}\t\t# {}".format(
                        ip, " ".join(host_info["aliases"]), host_info["comment"]
                    )
                else:
                    line = "{}\t\t{}".format(ip, " ".join(host_info["aliases"]))
        lines.append(line)

    hfn = _get_or_create_hostfile()
    with salt.utils.files.fopen(hfn, "w+") as ofile:
        for line in lines:
            if line.strip():
                # /etc/hosts needs to end with a newline so that some utils
                # that read it do not break
                ofile.write(
                    salt.utils.stringutils.to_str(line.strip() + str(os.linesep))
                )



"""
Docker Proxy Minion

.. versionadded:: 2019.2.0

:depends: docker

This proxy minion is just a shim to the docker executor, which will use the
:py:func:`docker.call <salt.modules.dockermod.call>` for everything except
state runs.


To configure the proxy minion:

.. code-block:: yaml

    proxy:
      proxytype: docker
      name: festive_leakey

It is also possible to just name the proxy minion the same name as the
container, and use grains to configure the proxy minion:

.. code-block:: yaml

    proxy:
        proxytype: docker
        name: {{grains['id']}}

name

    Name of the docker container
"""

__proxyenabled__ = ["docker"]
__virtualname__ = "docker"


def __virtual__():
    if __opts__.get("proxy", {}).get("proxytype") != __virtualname__:
        return False, f"Proxytype does not match: {__virtualname__}"
    return True


def module_executors():
    """
    List of module executors to use for this Proxy Minion
    """
    return [
        "docker",
    ]


def init(opts):
    """
    Always initialize
    """
    __context__["initialized"] = True


def initialized():
    """
    This should always be initialized
    """
    return __context__.get("initialized", False)


def shutdown(opts):
    """
    Nothing needs to be done to shutdown
    """
    __context__["initialized"] = False



"""
    Test cases for salt.modules.uswgi
"""

import pytest

import salt.modules.uwsgi as uwsgi
from tests.support.mock import MagicMock, Mock, patch


@pytest.fixture
def configure_loader_modules():
    with patch("salt.utils.path.which", Mock(return_value="/usr/bin/uwsgi")):
        return {uwsgi: {}}


def test_uwsgi_stats():
    socket = "127.0.0.1:5050"
    mock = MagicMock(return_value='{"a": 1, "b": 2}')
    with patch.dict(uwsgi.__salt__, {"cmd.run": mock}):
        result = uwsgi.stats(socket)
        mock.assert_called_once_with(
            ["uwsgi", "--connect-and-read", f"{socket}"],
            python_shell=False,
        )
        assert result == {"a": 1, "b": 2}



import logging

import pytest

log = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def non_blockdev_file():
    with pytest.helpers.temp_file() as tmp_file:
        assert tmp_file.is_file()
        yield tmp_file


def test_tuned_not_block_device(states, non_blockdev_file):
    """
    Test to ensure that when the target is not a block device,
    the state returns False and a comment indicating the issue.
    """
    name = str(non_blockdev_file)

    ret = states.blockdev.tuned(
        name=name,
    )
    assert ret["result"] is False
    assert ret["comment"] == f"Changes to {name} cannot be applied. Not a block device."



import pytest

import salt.auth.rest as rest
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    """
    Rest module configuration
    """
    return {
        rest: {
            "__opts__": {
                "external_auth": {
                    "rest": {"^url": "https://test_url/rest", "fred": [".*", "@runner"]}
                }
            }
        }
    }


def test_rest_auth_config():
    ret = rest._rest_auth_setup()
    assert ret == "https://test_url/rest"


def test_fetch_call_failed():
    with patch("salt.utils.http.query", MagicMock(return_value={"status": 401})):
        ret = rest.fetch("foo", None)
        assert ret is False


def test_fetch_call_success_dict_none():
    with patch(
        "salt.utils.http.query", MagicMock(return_value={"status": 200, "dict": None})
    ):
        ret = rest.fetch("foo", None)
        assert ret == []


def test_fetch_call_success_dict_acl():
    with patch(
        "salt.utils.http.query",
        MagicMock(return_value={"status": 200, "dict": {"foo": ["@wheel"]}}),
    ):
        ret = rest.fetch("foo", None)
        assert ret == {"foo": ["@wheel"]}


def test_auth_nopass():
    ret = rest.auth("foo", None)
    assert ret is False


def test_auth_nouser():
    ret = rest.auth(None, "foo")
    assert ret is False


def test_auth_nouserandpass():
    ret = rest.auth(None, None)
    assert ret is False


def test_auth_ok():
    with patch(
        "salt.utils.http.query",
        MagicMock(return_value={"status": 200, "dict": ["@wheel"]}),
    ):
        ret = rest.auth("foo", None)
        assert ret is True


def test_acl_without_merge():
    ret = rest.acl("fred", password="password")
    assert ret == [".*", "@runner"]


def test_acl_unauthorized():
    with patch("salt.utils.http.query", MagicMock(return_value={"status": 400})):
        ret = rest.acl("foo", password="password")
        assert ret is None


def test_acl_no_merge():
    with patch(
        "salt.utils.http.query", MagicMock(return_value={"status": 200, "dict": None})
    ):
        ret = rest.acl("fred", password="password")
        assert ret == [".*", "@runner"]


def test_acl_merge():
    with patch(
        "salt.utils.http.query",
        MagicMock(return_value={"status": 200, "dict": ["@wheel"]}),
    ):
        ret = rest.acl("fred", password="password")
        assert ret == [".*", "@runner", "@wheel"]
